/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "Lexer.hpp"
#include "Parser.hpp"

#include "Cell.hpp"

#include <cctype>
#include <cassert>

namespace lispel
{
namespace depel
{

/* ************************************************************************** */
/* ********************************** Lexer ********************************* */
/* ************************************************************************** */

int Lexer::yylex(semanticT& yylval)
{
  assert(in);

  skip();
  if(in->eof())
    return yy::parser::token::EOF_SYMBOL;

  // Check next character to see what comes
  char c(in->get());
  assert(!in->eof()); // Would be caught above

  // Cell-opening, we don't allow .xx numbers!
  if(c=='"' || c=='[' || std::isdigit(c) || c=='-')
  {
    const bool realCell(c=='[');
    if(!realCell)
      in->unget();
    yylval.cell=new Cell(*in);
    if(realCell)
    {
      skip();
      c=in->get();
      if(c!=']')
        throw std::runtime_error("Expected ']'!");
    }
    return yy::parser::token::CELL;
  }

  // Atom? Can only start with alphanum or _
  if(std::isalnum(c) || c=='_')
  {
    in->unget();
    Cell c(*in);
    assert(c.getType()==Cell::ATOM);
    return procAtom(yylval, c.getAtom());
  }

  // Return character as-is otherwise
  return c;
}

int Lexer::procAtom(semanticT& yylval, const Atom& a)
{
  // Check keywords
  #define CHECK_KEYWORD(name, str) \
    if(a==Atom(str)) \
      return yy::parser::token::name;
  CHECK_KEYWORD(VON, "von")
  CHECK_KEYWORD(UND, "und")
  CHECK_KEYWORD(ALS, "als")
  CHECK_KEYWORD(SSEI, "Sei")
  CHECK_KEYWORD(SEI, "sei")
  CHECK_KEYWORD(IST, "ist")
  CHECK_KEYWORD(HAT, "hat")
  CHECK_KEYWORD(FUER, "fuer")
  CHECK_KEYWORD(SONST, "sonst")
  CHECK_KEYWORD(MIT, "mit")
  CHECK_KEYWORD(BEDEUTET, "bedeutet")
  CHECK_KEYWORD(MACHE, "mache")

  // Check article
  static const char* der[]={"der", "die", "das", "dem", "der", "des", "den"};
  static const char* Der[]={"Der", "Die", "Das"};
  static const char* ein[]={"ein", "eine", "einem", "einer"};
  static const char* Ein[]={"Ein", "Eine"};
  #define CHECK_ARTICLE(name, list) \
    for(int i(0); i!=sizeof(list)/sizeof(list[0]); ++i) \
      if(a==Atom(list[i])) \
        return yy::parser::token::name;
  CHECK_ARTICLE(DER, der)
  CHECK_ARTICLE(EIN, ein)
  CHECK_ARTICLE(DDER, Der)
  CHECK_ARTICLE(EEIN, Ein)

  // Atom or operator
  yylval.atom=new Atom(a);
  return operators.find(a)!=operators.end() ?
    yy::parser::token::OPERATOR :
    yy::parser::token::ATOM;
}

void Lexer::skip()
{
  char c;
  while(c=in->get(), std::isspace(c) || skipComment(c));
  unget();
}

bool Lexer::skipComment(char c)
{
  // Must be '/' for a comment!
  if(c!='/')
    return false;

  // Check next character
  c=in->get();
  switch(c)
  {

    // C-style multi-line comment
    case '*':
      char last;
      do
      {
        last=c;
        c=in->get();
      } while((last!='*' || c!='/') && *in);
      return true;

    // C++-style single-line comment
    case '/':
      while(c=in->get(), c!='\n' && *in);
      return true;

    // No comment at all
    default:
      unget(); // Unget the second character read
      return false;

  }

  assert(false);
}

} // namespace depel
} // namespace lispel
