/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef DEPEL_LEXER_HPP
#define DEPEL_LEXER_HPP

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "Parser.tab.hh"

#include "Atom.hpp"

#include <iostream>
#include <set>

namespace lispel
{
namespace depel
{

/* ************************************************************************** */
/* ********************************** Lexer ********************************* */
/* ************************************************************************** */

/**
 * The lexer class for parsing.  The lexer identifies Cells, keywords for de-pel
 * (like der/die/das, einer/eine/ein, von, und) and punctuation; it is designed
 * to work together with bison.
 */
class Lexer
{

  /* ************************************************************************ */
  public:

    /** The type of the semantic token */
    typedef yy::parser::semantic_type semanticT;

    /**
     * Construct the lexer, without input stream at first.
     */
    inline Lexer()
      : in(NULL)
    {}

    // No destructor, stream is not owned by us

    /**
     * Set the input stream
     * @param i The stream to read from
     */
    inline void setInput(std::istream& i)
    {
      in=&i;
    }

    /**
     * Define an Atom to be an operator
     * @param a The atom to define as operator
     */
    inline void defineOperator(const Atom& a)
    {
      operators.insert(a);
    }

    /**
     * The lex-function itself, called by a wrapper for bison.
     * @param yylval Store the semantic value here
     * @return The token type
     */
    int yylex(semanticT& yylval);

  /* ************************************************************************ */
  private:

    /** The input stream read from, not owned */
    std::istream* in;

    /** The operator-array type */
    typedef std::set<Atom> operatorSet;
    /** The defined operators */
    operatorSet operators;

    /**
     * Process an Atom-value, testing for keywords
     * @param yylval Return non-keyword value here
     * @param a The Atom value
     * @return The token ID
     */
    int procAtom(semanticT& yylval, const Atom& a);

    /**
     * Skip the appearing white space including comments.
     */
    void skip();

    /**
     * Skip appearing comment, if any
     * @param c The first character of the part to skip
     * @return True if there was a comment skipped
     */
    bool skipComment(char c);

    /**
     * Unget the last character (including EOF-checking)
     */
    inline void unget()
    {
      if(!in->eof())
        in->unget();
    }

};

} // namespace depel
} // namespace lispel

#endif // Header-guard
