/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef DEPEL_PARSER_HPP
#define DEPEL_PARSER_HPP

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "Lexer.hpp"
#include "Parser.tab.hh"

#include "Cell.hpp"

#include <iostream>
#include <stdexcept>

namespace lispel
{
namespace depel
{

/* ************************************************************************** */
/* ********************************* Parser ********************************* */
/* ************************************************************************** */

/**
 * The parser interface used in the code, it encapsulates the bison parser.
 */
class Parser
{

  /* ************************************************************************ */
  public:

    /**
     * EOF exception thrown
     */
    class EndOfInput : public std::runtime_error
    {
      public:
        inline EndOfInput()
          : std::runtime_error("EOF")
        {}
    };

    /**
     * Construct the parser, still without input stream; one needs to be
     * supplied later.
     */
    inline Parser()
      : lex(), bison(*this, lex), value(NULL)
    {}

    /**
     * Destroy, freeing the hold value.
     */
    ~Parser();

    /**
     * Set the input stream to use
     * @param in The stream to use
     */
    inline void setInput(std::istream& in)
    {
      lex.setInput(in);
    }

    /**
     * Parse a declaration / action to process.
     * @return The parsed declaration, must be freed
     * @throws EndOfInput if end of input reached
     */
    Cell* parseDeclaration();

    friend class yy::parser;

  /* ************************************************************************ */
  private:

    /** The Lexer used */
    Lexer lex;

    /** The bison parser used internally */
    yy::parser bison;

    /** The value hold, stored here by the parser */
    Cell* value;

    /**
     * Set the currently hold value.
     * @param v The new value
     */
    inline void setValue(Cell* v =NULL)
    {
      if(value)
        delete value;
      value=v;
    }

};

} // namespace depel
} // namespace lispel

#endif // Header-guard
