%language "C++"
%defines
%error-verbose
%parse-param {Parser& p}
%parse-param {Lexer& l}
%lex-param {Lexer& l}
%expect 9

%code requires {

#include "Cell.hpp"
#include "Atom.hpp"
#include "List.hpp"

namespace lispel
{
namespace depel
{
  class Parser;
  class Lexer;
}
}

using namespace lispel;
using namespace ellisp;
using namespace depel;

} /* code requires */

/* The semantic value type */
%union
{
  /** Cell-value */
  Cell* cell;
  /** Atom-value */
  Atom* atom;
  /** List-value */
  List* list;
}

/* Token types for Atoms and Cells from lexer */
%token <atom> ATOM
%token <cell> CELL

/* Already defined operators are recognized by lexer, too */
%token <atom> OPERATOR

/* Keyword tokens */
%token DER
%token DDER
%token EIN
%token EEIN
%token VON
%token UND
%token ALS
%token SSEI
%token SEI
%token IST
%token HAT
%token FUER
%token SONST
%token MIT
%token BEDEUTET
%token MACHE

/* End of input token */
%token EOF_SYMBOL 0

/* Types for values and declarations are cells */
%type <cell> name
%type <cell> innerValue
%type <cell> value
%type <cell> declaration

/* Whole result of the parsing */
%type <cell> result

/* Argument-lists are build as lists */
%type <list> funcarglist
%type <list> comma_funcarglist
%type <cell> def_funcarglist
%type <list> comma_argdef
%type <cell> argdef

/* Super-class and member lists */
%type <cell> member_list
%type <list> comma_memberlist
%type <cell> memberdef
%type <cell> super_list
%type <list> comma_superlist
%type <cell> superdef

/* Action literals and actions */
%type <list> actioncall
%type <list> lazySequence
%type <list> actarglist
%type <cell> def_actarglist
%type <list> actbodyStmt
%type <list> actbody

/* Destructor is simply free it */
%destructor
{
  delete $$;
} ATOM CELL OPERATOR value innerValue declaration result
%destructor
{
  delete $$;
} funcarglist comma_funcarglist def_funcarglist argdef comma_argdef
%destructor
{
  delete $$;
} super_list comma_superlist superdef member_list comma_memberlist memberdef
%destructor
{
  delete $$;
} actioncall lazySequence actarglist def_actarglist actbodyStmt actbody

/* Only for .cc file, not header */
%code {

#include "Parser.hpp"
#include "Lexer.hpp"

#include "Atom.hpp"

int yylex(Lexer::semanticT* yylval, Lexer& l);
static Cell* operatorName(const Atom& a);

} /* code */

%start start

/* ************************************************************************** */
%%

/* Wrap raw atoms not being keywords into cells */
name:
  ATOM
  {
    $$=new Cell(*$1);
    delete $1; /* Bison discards the symbol for the above */
  }
;

/* ******************************** Values ********************************** */

/* A value not consisting of lists of other values */
innerValue:
  /* A single Atom may be a value */
  name
  {
    $$=$1;
  }
  /* A cell in general is a value */
| CELL
  {
    $$=$1;
  }
  /* Simple cast */
| innerValue ALS name
  {
    $$=new Cell(List::build(new Cell(Atom("_cast")), $3, $1, NULL));
  }
;

/* Fully fledged value, including those built of others by function-calls */
value:
  /* Simply an inner value */
  innerValue
  {
    $$=$1;
  }
  /* Function call with argument list */
| DER name funcarglist
  {
    $$=new Cell(new List($2, $3));
  }
  /* Function call cast */
| DER name ALS name funcarglist
  {
    $$=new Cell(List::build(new Cell(Atom("_cast")), $4,
      new Cell(new List($2, $5)), NULL));
  }
  /* Conditional value */
| FUER value value ',' SONST value
  {
    $$=new Cell(List::build(new Cell(Atom("_if")), $2, $3, $6, NULL));
  }
  /* Unary operator, need not operate on innerValue! */
| OPERATOR value
  {
    $$=new Cell(List::build(operatorName(*$1), $2, NULL));
    delete $1;
  }
  /* Binary operator, only first one needs to be innerValue */
| innerValue OPERATOR value
  {
    $$=new Cell(List::build(operatorName(*$2), $1, $3, NULL));
    delete $2;
  }
  /* Literal action */
| actioncall
  {
    $$=new Cell(new List(new Cell(Atom("!")), $1));
  }
  /* Lazy action sequence */
| lazySequence
  {
    $$=new Cell(new List(new Cell(Atom("_sequence")), $1));
  }
;

/* ******************************* Function arguments *********************** */

funcarglist:
  value
  {
    $$=new List($1);
  }
| VON comma_funcarglist UND value
  {
    $2->pushBack($4);
    $$=$2;
  }
;

comma_funcarglist:
  value
  {
    $$=new List($1);
  }
| comma_funcarglist ',' value
  {
    $$=$1;
    $$->pushBack($3);
  }
;

def_funcarglist:
  VON argdef
  {
    $$=new Cell(new List($2));
  }
| VON comma_argdef UND argdef
  {
    $2->pushBack($4);
    $$=new Cell($2);
  }
;

comma_argdef:
  argdef
  {
    $$=new List($1);
  }
| comma_argdef ',' argdef
  {
    $1->pushBack($3);
    $$=$1;
  }
;

argdef:
  DER name name
  {
    $$=new Cell(List::build($2, $3, NULL));
  }
;

/* **************************** Super- and member lists ********************* */

member_list:
  HAT memberdef
  {
    $$=new Cell(new List($2));
  }
| HAT comma_memberlist UND memberdef
  {
    $2->pushBack($4);
    $$=new Cell($2);
  }
;

comma_memberlist:
  memberdef
  {
    $$=new List($1);
  }
| comma_memberlist ',' memberdef
  {
    $1->pushBack($3);
    $$=$1;
  }
;

memberdef:
  DER name name
  {
    $$=new Cell(List::build($2, $3, NULL));
  }
;

super_list:
  IST superdef
  {
    $$=new Cell(List::build($2, NULL));
  }
| IST comma_superlist UND superdef
  {
    $2->pushBack($4);
    $$=new Cell($2);
  }
;

comma_superlist:
  superdef
  {
    $$=new List($1);
  }
| comma_superlist ',' superdef
  {
    $1->pushBack($3);
    $$=$1;
  }
;

superdef:
  /* Without given name */
  EIN name
  {
    $$=new Cell(new List($2));
  }
| /* With given name */
  EIN name '(' DER name ')'
  {
    $$=new Cell(List::build($2, $5, NULL));
  }
;

/* ********************************* Actions ******************************** */

actarglist:
  MIT value
  {
    $$=new List($2);
  }
| MIT comma_funcarglist UND value
  {
    $2->pushBack($4);
    $$=$2;
  }
;

actioncall:
  /* Action without arguments, not allowed */
  /*
  name '!'
  {
    $$=new List($1);
  }
  */
  /* Action with arguments */
  name actarglist '!'
  {
    $$=new List($1, $2);
  }
  /* Do-call of value containing action */
| MACHE value '!'
  {
    $$=List::build(new Cell(Atom("_do")), $2, NULL);
  }
;

def_actarglist:
  MIT argdef
  {
    $$=new Cell(List::build($2, NULL));
  }
| MIT comma_argdef UND argdef
  {
    $2->pushBack($4);
    $$=new Cell($2);
  }
;

actbodyStmt:
  actioncall
  {
    $$=$1;
  }
| SEI name value '!'
  {
    $$=List::build(new Cell(Atom("_constant")), $2, $3, NULL);
  }
;

actbody:
  actbodyStmt
  {
    $$=new List(new Cell($1));
  }
| actbody actbodyStmt
  {
    $1->pushBack(new Cell($2));
    $$=$1;
  }
;

lazySequence:
  value '&' value
  {
    $$=List::build($1, $3, NULL);
  }
| lazySequence '&' value
  {
    $1->pushBack($3);
    $$=$1;
  }
;

/* ***************************** Put it all together ************************ */

declaration:
  /* Constant definition */
  SSEI name value '.'
  {
    $$=new Cell(List::build(new Cell(Atom("_constant")), $2, $3, NULL));
  }
  /* Function definition */
| DDER name def_funcarglist IST value '.'
  {
    $$=new Cell(List::build(new Cell(Atom("_function")), $2, $3, $5, NULL));
  }
  /* Class definition, only members */
| EEIN name member_list '.'
  {
    $$=new Cell(List::build(new Cell(Atom("_class")), $2,
      new Cell(List::build()), $3, NULL));
  }
  /* Class definition, only supers */
| EEIN name super_list '.'
  {
    $$=new Cell(List::build(new Cell(Atom("_class")), $2,
      $3, new Cell(List::build()), NULL));
  }
  /* Class definition, supers and members */
| EEIN name super_list ';' UND member_list '.'
  {
    $$=new Cell(List::build(new Cell(Atom("_class")), $2, $3, $6, NULL));
  }
  /* Definition of unary operator: operator type arg IST body */
| name name name IST value '.'
  {
    l.defineOperator($1->getAtom());
    $$=new Cell(List::build(new Cell(Atom("_function")),
      operatorName($1->getAtom()),
      new Cell(new List(new Cell(List::build($2, $3, NULL)))),
      $5, NULL));
    delete $1;
  }
  /* Definition of binary operator: type arg1 operator type arg2 IST body */
| name name name name name IST value '.'
  {
    l.defineOperator($3->getAtom());
    $$=new Cell(List::build(new Cell(Atom("_function")),
      operatorName($3->getAtom()),
      new Cell(List::build(
        new Cell(List::build($1, $2, NULL)),
        new Cell(List::build($4, $5, NULL)),
        NULL)),
      $7, NULL));
    delete $3;
  }
  /* Definition of action without arguments, not allowed */
  /*
| name BEDEUTET ':' actbody '.'
  {
    $$=new Cell(List::build(new Cell(Atom("_action")), $1,
      new Cell(List::build()),
      new Cell($4), NULL));
  }
  */
  /* Definition of action with arguments */
| name BEDEUTET def_actarglist ':' actbody '.'
  {
    $$=new Cell(List::build(new Cell(Atom("_action")), $1,
      $3, new Cell($5), NULL));
  }
;

result:
  declaration
  {
    $$=$1;
  }
| value '?'
  {
    $$=$1;
  }
| actioncall '!' /* This makes !!, one from the action-call */
  {
    $$=new Cell($1);
  }
| EOF_SYMBOL
  {
    // Empty value signals EOF
    $$=NULL;
  }
;

start:
  result
  {
    p.setValue($1);
    YYACCEPT;
  }
;

/* ************************************************************************** */
%%

#include <iostream>
#include <string>

// Redirect yylex
int yylex(Lexer::semanticT* yylval, Lexer& l)
{
  const int token(l.yylex(*yylval));
  //std::cout << token << " " << (char)token << std::endl;
  return token;
}

// Build name of operator-function
Cell* operatorName(const Atom& a)
{
  return new Cell(Atom("_operator_"+a.str()));
}

void yy::parser::error(const location_type& loc, const std::string& msg)
{
  std::cerr << "Parse error: " << msg << std::endl;
}
