/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008,2012  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#ifndef DATADIR
# warning DATADIR undefined, no library will be loaded at start-up!
#endif

#include "DepelShell.hpp"

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>

#ifdef GETOPT
# include <getopt.h>
#endif

using namespace lispel;
using namespace depel;

using namespace std;

/**
 * Print version information
 */
void printVersion()
{
  cout << "This is depel of Lispel " VERSION << "." << endl;
  cout << "Copyright (C) 2008,2012 by Daniel Kraft" << endl;
  cout << "Released under the GNU General Public License, version 3" << endl;
  cout << endl;
}

/**
 * Print help information depending on getopt-ness
 */
void printHelp()
{
#ifndef GETOPT
  cout << "Usage: depel [FILE]" << endl;
  cout << "If specified, process FILE instead of interactive shell." << endl;
#else /* getopt */
  cout << "Usage: depel [OPTIONS] [FILE]" << endl;
  cout << "If specified, process FILE instead of interactive shell." << endl;
  cout << endl << "OPTIONS:" << endl;
  cout << "  --help     print this help message" << endl;
  cout << "  --version  print version information" << endl;
  cout << "  --batch    start non-interactive, always if FILE is given" << endl;
#endif /* getopt */

  cout << endl;
}

int main(int argc, char** argv)
{
  std::istream* in(&cin); // This is set to the input-stream to read from
  bool ourIn(false); // Close and delete the input-stream object afterwards?

  // Use int's for getopt
  int help(0), version(0); // Print help/version information?
  int interact(1); // Start in interactive mode?

#ifndef GETOPT
  // Without getopt, only recognize single filename as input if given
  if(argc>2)
  {
    printHelp();
    return EXIT_FAILURE;
  }
  if(argc==2)
  {
    interact=0;
    outIn=true;
    in=new ifstream(argv[1]);
  }
#else /* getopt */
  // Use getopt to parse the arguments

  // Define the option-structure
  static const option long_options[]={
    {"help", no_argument, &help, 1},
    {"version", no_argument, &version, 1},
    {"batch", no_argument, &interact, 0},
    {0, 0, 0, 0}};
  int option_index(0);

  // Read the options
  int c;
  do
    c=getopt_long(argc, argv, "", long_options, &option_index);
  while(c!=-1);

  // Finally, see if file is given too
  if(optind+1<argc)
  {
    printHelp();
    return EXIT_FAILURE;
  }
  if(optind<argc)
  {
    interact=0;
    ourIn=true;
    in=new ifstream(argv[optind++]);
  }
  assert(optind==argc);

#endif /* getopt */

  // Print help/version as requested and exit
  if(version)
    printVersion();
  if(help)
    printHelp();
  if(version || help)
    return EXIT_SUCCESS;

  // Print welcome message if interactive
  if(interact)
    printVersion();

  // Create the shell
  assert(in);
  DepelShell shell(*in, cout, interact);

  // Load library
#ifdef DATADIR
  {
    std::ifstream libIn(DATADIR "/depel.dpl");
    if(!libIn)
      cerr << "Warning:  Could not open library!" << endl;
    else
      shell.include(libIn);
    libIn.close();
  }
#endif

  // Start main processing
  shell.run();

  // Print final newline
  if(interact)
    cout << endl;

  if(ourIn)
  {
    static_cast<std::ifstream*>(in)->close();
    delete in;
  }

  return EXIT_SUCCESS;
}
