/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "Arithmetic.hpp"

#include "Elob.hpp"
#include "Runtime.hpp"
#include "Atom.hpp"
#include "Function.hpp"

namespace lispel
{
namespace ellib
{

namespace Arithmetic
{

#define NAME_ADD Atom("Summe")
#define NAME_SUBTR Atom("Differenz")
#define NAME_MUL Atom("Produkt")
#define NAME_DIV Atom("Quotient")
#define NAME_MOD Atom("Rest")

void define(Runtime& rt)
{
  rt.addFunction(NAME_ADD,
    new Cummulative(rt,
      0, &Cummulative::intUpdateSum,
      0., &Cummulative::numUpdateSum));
  rt.addFunction(NAME_MUL,
    new Cummulative(rt,
      1, &Cummulative::intUpdateProd,
      1., &Cummulative::numUpdateProd));

  rt.addFunction(NAME_SUBTR,
    new NonCummulative(rt,
      &NonCummulative::intSubtr,
      &NonCummulative::numSubtr));
  rt.addFunction(NAME_DIV,
    new NonCummulative(rt,
      &NonCummulative::intDiv,
      &NonCummulative::numDiv));

  // Modulo only takes int, int
  Signature modSig;
  modSig.push_back(rt.getIntegerType());
  modSig.push_back(rt.getIntegerType());
  rt.addFunction(NAME_MOD, new CallbackFunction(rt, modSig, &modulo));
}

/* ************************************************************************** */
/* ******************************* Cummulative ****************************** */
/* ************************************************************************** */

bool Cummulative::accepts(const Arguments& args) const
{
  // We want 1+ arguments
  if(args.empty())
    return false;

  // Now, we need things conversible to numbers
  Type::Ref numType(rt.getNumberType());
  for(Arguments::const_iterator i(args.begin()); i!=args.end(); ++i)
    if(!numType->isAssignableFrom(i->getType()))
      return false;

  return true;
}

Elob Cummulative::eval(const Arguments& args) const
{
  assert(!args.empty());

  Type::Ref intType(rt.getIntegerType());
  Type::Ref numType(rt.getNumberType());

  /* Check argument-types, run the sum.  This is done with a sum for integers
   * and one for numbers; in the end, if not only integers, those are summed.
   */
  bool onlyInt(true);
  Cell::intType intVal(intStart);
  Cell::numberType numVal(numStart);
  for(Arguments::const_iterator i(args.begin()); i!=args.end(); ++i)
  {
    if(intType->isAssignableFrom(i->getType()))
      intUp(intVal, i->castTo(intType).getInteger());
    else
    {
      assert(numType->isAssignableFrom(i->getType()));
      numUp(numVal, i->castTo(numType).getNumber());
      onlyInt=false;
    }
  }

  // Only integers?
  if(onlyInt)
  {
    assert(numVal==numStart);
    return Elob::forInteger(intVal, rt);
  }

  // No, number result.
  numUp(numVal, static_cast<Cell::numberType>(intVal));
  return Elob::forNumber(numVal, rt);
}

void Cummulative::intUpdateSum(Cell::intType& dest, Cell::intType inc)
{
  dest+=inc;
}

void Cummulative::numUpdateSum(Cell::numberType& dest, Cell::numberType inc)
{
  dest+=inc;
}

void Cummulative::intUpdateProd(Cell::intType& dest, Cell::intType inc)
{
  dest*=inc;
}

void Cummulative::numUpdateProd(Cell::numberType& dest, Cell::numberType inc)
{
  dest*=inc;
}

/* ************************************************************************** */
/* ******************************** NonCummulative ************************** */
/* ************************************************************************** */

bool NonCummulative::accepts(const Arguments& args) const
{
  // We want exactly two arguments
  if(args.size()!=2)
    return false;

  // Both must be numbers
  Type::Ref numType(rt.getNumberType());
  if(!numType->isAssignableFrom(args.front().getType()))
    return false;
  if(!numType->isAssignableFrom(args.back().getType()))
    return false;

  return true;
}

Elob NonCummulative::eval(const Arguments& args) const
{
  assert(args.size()==2);

  Type::Ref intType(rt.getIntegerType());
  Type::Ref numType(rt.getNumberType());

  // Both integers?
  if(intType->isAssignableFrom(args.front().getType()) &&
     intType->isAssignableFrom(args.back().getType()))
  {
    const Cell::intType val1(args.front().castTo(intType).getInteger());
    const Cell::intType val2(args.back().castTo(intType).getInteger());
    return Elob::forInteger(intC(val1, val2), rt);
  }

  // No, must be numbers
  const Cell::numberType val1(args.front().castTo(numType).getNumber());
  const Cell::numberType val2(args.back().castTo(numType).getNumber());
  return Elob::forNumber(numC(val1, val2), rt);
}

Cell::intType NonCummulative::intSubtr(Cell::intType a, Cell::intType b)
{
  return a-b;
}

Cell::numberType NonCummulative::numSubtr(
  Cell::numberType a, Cell::numberType b)
{
  return a-b;
}

Cell::intType NonCummulative::intDiv(Cell::intType a, Cell::intType b)
{
  return a/b;
}

Cell::numberType NonCummulative::numDiv(Cell::numberType a, Cell::numberType b)
{
  return a/b;
}

/* ************************************************************************** */
/* ******************************* Modulo *********************************** */
/* ************************************************************************** */

Elob modulo(const Runtime& rt, const Arguments& args)
{
  assert(args.size()==2);
  assert(args.front().getType()==rt.getIntegerType());
  assert(args.back().getType()==rt.getIntegerType());

  return Elob::forInteger(args.front().getInteger()%args.back().getInteger(),
    rt);
}

} // namespace Arithmetic

} // namespace ellib
} // namespace lispel
