/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef ELLIB_ARITHMETIC_HPP
#define ELLIB_ARITHMETIC_HPP

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "Function.hpp"

namespace lispel
{
namespace ellisp
{
  class Runtime;
}
using namespace ellisp;
namespace ellib
{

/**
 * Functions for basic arithmetic.
 */
namespace Arithmetic
{

  /**
   * Register those functions with a Runtime
   */
  void define(Runtime& rt);

  /* ************************************************************************ */
  /* ***************************** Cummulative ****************************** */
  /* ************************************************************************ */

  /**
   * Function-class doing cummulative arithmetic, sum and product.
   */
  class Cummulative : public Function
  {

    /* ********************************************************************** */
    public:

      /**
       * Type for callback updating integer with new one
       * @param dest The value to update
       * @param inc The new value to "cummulate" to it
       */
      typedef void (*IntUpdate)(Cell::intType& dest, Cell::intType inc);

      /**
       * @see IntUpdate
       */
      typedef void (*NumUpdate)(Cell::numberType& dest, Cell::numberType inc);

      /**
       * Construct it, given the start values and updaters for int/num
       * @param r The runtime
       * @param is Int start
       * @param iu Int updater
       * @param ns Num start
       * @param nu Num updater
       */
      inline Cummulative(const Runtime& r,
        Cell::intType is, IntUpdate iu,
        Cell::numberType ns, NumUpdate nu)
        : Function(r),
          intStart(is), intUp(iu),
          numStart(ns), numUp(nu)
      {}

      /**
       * @see Function::accepts
       */
      virtual bool accepts(const Arguments& args) const;

      /**
       * @see Function::eval
       */
      virtual Elob eval(const Arguments& args) const;

      // Updaters for sum/product
      static void intUpdateSum(Cell::intType& dest, Cell::intType inc);
      static void numUpdateSum(Cell::numberType& dest, Cell::numberType inc);
      static void intUpdateProd(Cell::intType& dest, Cell::intType inc);
      static void numUpdateProd(Cell::numberType& dest, Cell::numberType inc);

    /* ********************************************************************** */
    private:

      /** Start value for integers */
      const Cell::intType intStart;

      /** Updater for integers */
      const IntUpdate intUp;

      /** Start value for numbers */
      const Cell::numberType numStart;

      /** Updater for numbers */
      const NumUpdate numUp;

  };

  /* ************************************************************************ */
  /* ****************************** NonCummulative ************************** */
  /* ************************************************************************ */

  /**
   * Function-class doing non-cummulative real arithmetic, subtract and divide.
   */
  class NonCummulative : public Function
  {

    /* ********************************************************************** */
    public:

      /**
       * Type for callback calculating integers with new one
       * @param a First operand
       * @param b Second operand
       */
      typedef Cell::intType (*IntCalc)(Cell::intType a, Cell::intType b);

      /**
       * @see IntCalc
       */
      typedef Cell::numberType (*NumCalc)(
        Cell::numberType a,
        Cell::numberType b);

      /**
       * Construct it, given the calculators
       * @param r The runtime
       * @param i The int calculator
       * @param n The num calculator
       */
      inline NonCummulative(const Runtime& r, IntCalc i, NumCalc n)
        : Function(r), intC(i), numC(n)
      {}

      /**
       * @see Function::accepts
       */
      virtual bool accepts(const Arguments& args) const;

      /**
       * @see Function::eval
       */
      virtual Elob eval(const Arguments& args) const;

      // Calculators for subtraction/division
      static Cell::intType intSubtr(Cell::intType a, Cell::intType b);
      static Cell::numberType numSubtr(Cell::numberType a, Cell::numberType b);
      static Cell::intType intDiv(Cell::intType a, Cell::intType b);
      static Cell::numberType numDiv(Cell::numberType a, Cell::numberType b);

    /* ********************************************************************** */
    private:

      /** Calculator for integers */
      const IntCalc intC;

      /** Calculator for numbers */
      const NumCalc numC;

  };

  /* ************************************************************************ */
  /* ***************************** Modulo *********************************** */
  /* ************************************************************************ */

  /* Modulo is done by CallbackFunction; it accepts only int arguments */
  Elob modulo(const Runtime& rt, const Arguments& args);

} // namespace Arithmetic

} // namespace ellib
} // namespace lispel

#endif // Header-guard
