/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "CellTypes.hpp"

#include "Type.hpp"
#include "Runtime.hpp"
#include "Atom.hpp"
#include "Cell.hpp"

#include <string>

namespace lispel
{
namespace ellib
{

namespace CellTypes
{

#define NAME_INTEGER Atom("Ganzzahl")
#define NAME_NUMBER Atom("Zahl")
#define NAME_STRING Atom("String")

void define(Runtime& rt)
{
  rt.addType(NAME_NUMBER,
    new TopType<Cell::numberType>(NAME_NUMBER, rt));
  rt.addType(NAME_INTEGER, new IntType(rt));
  rt.addType(NAME_STRING, new TopType<std::string>(NAME_STRING, rt));
}

/* ******************************* IntType ********************************** */

IntType::IntType(const Runtime& r)
  : Type(NAME_INTEGER, r)
{
  addSuper(rt.getNumberType());
}

void IntType::destroyData(void* val) const
{
  delete static_cast<Cell::intType*>(val);
}

Elob IntType::castValueTo(const Elob& val, Ref t) const
{
  assert(val.getType()==ref());

  // Might only be casted to number
  if(t!=rt.getNumberType())
  {
    std::ostringstream msg;
    msg << getName().str() << " can only be casted to number, not ";
    msg << t->getName().str();
    throw std::runtime_error(msg.str());
  }

  const Cell::intType& intVal(*static_cast<const Cell::intType*>(
    val.getValue()));

  return Elob::forNumber(static_cast<Cell::numberType>(intVal), rt);
}

std::ostream& IntType::print(std::ostream& out, const void* val) const
{
  out << *static_cast<const Cell::intType*>(val);
  return out;
}

} // namespace CellTypes

} // namespace ellib

namespace ellisp
{

using namespace ellib;
using namespace CellTypes;

/* ******************************* Runtime ********************************** */

Type::Ref Runtime::getIntegerType() const
{
  return queryType(NAME_INTEGER);
}

Type::Ref Runtime::getNumberType() const
{
  return queryType(NAME_NUMBER);
}

Type::Ref Runtime::getStringType() const
{
  return queryType(NAME_STRING);
}

/* ********************************** Elob ********************************** */

Cell::intType Elob::getInteger() const
{
  return *static_cast<const Cell::intType*>(getValue());
}

Cell::numberType Elob::getNumber() const
{
  return *static_cast<const Cell::numberType*>(getValue());
}

const std::string& Elob::getString() const
{
  return *static_cast<const std::string*>(getValue());
}

Elob Elob::forInteger(Cell::intType val, const Runtime& rt)
{
  return Elob(rt.getIntegerType(), new Cell::intType(val), rt);
}

Elob Elob::forNumber(Cell::numberType val, const Runtime& rt)
{
  return Elob(rt.getNumberType(), new Cell::numberType(val), rt);
}

Elob Elob::forString(const std::string& val, const Runtime& rt)
{
  return Elob(rt.getStringType(), new std::string(val), rt);
}

} // namespace ellisp
} // namespace lispel
