/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "Comparisons.hpp"

#include "Runtime.hpp"
#include "Elob.hpp"
#include "Atom.hpp"
#include "Cell.hpp"

#include <sstream>
#include <stdexcept>
#include <memory>

namespace lispel
{
namespace ellib
{

namespace Comparisons
{

#define NAME_EQUAL Atom("Gleichheit")
#define NAME_NOTEQUAL Atom("Ungleichheit")
#define NAME_LESSTHAN Atom("KleinerAls")
#define NAME_GREATERTHAN Atom("GroeszerAls")
#define NAME_LESSEQUAL Atom("KleinerGleich")
#define NAME_GREATEREQUAL Atom("GroeszerGleich")

void define(Runtime& rt)
{
  /* Generic operators */

  // a!=b <=> !(a==b): use ==, no swap, inversion
  rt.addFunction(NAME_NOTEQUAL,
    new Generic(rt, NAME_EQUAL, false, true));

  // a>b <=> b<a: use <, swap, no inversion
  rt.addFunction(NAME_GREATERTHAN,
    new Generic(rt, NAME_LESSTHAN, true, false));
  
  // a<=b <=> !(a>b) <=> !(b<a): use <, swap, inversion
  rt.addFunction(NAME_LESSEQUAL,
    new Generic(rt, NAME_LESSTHAN, true, true));

  // a>=b <=> !(a<b): use <, no swap, inversion
  rt.addFunction(NAME_GREATEREQUAL,
    new Generic(rt, NAME_LESSTHAN, false, true));

  /* Concrete definitions */

  rt.addFunction(NAME_EQUAL, new NumberComp(rt, true));
  rt.addFunction(NAME_LESSTHAN, new NumberComp(rt, false));

  Signature sig;
  sig.push_back(rt.getStringType());
  sig.push_back(rt.getStringType());
  rt.addFunction(NAME_EQUAL,
    new CallbackFunction(rt, sig, &stringEqual));
  rt.addFunction(NAME_LESSTHAN,
    new CallbackFunction(rt, sig, &stringLessThan));

  sig.clear();
  sig.push_back(rt.getBooleanType());
  sig.push_back(rt.getBooleanType());
  rt.addFunction(NAME_EQUAL,
    new CallbackFunction(rt, sig, &boolEqual));
}

Atom getEqualName()
{
  return NAME_EQUAL;
}

Atom getLessThanName()
{
  return NAME_LESSTHAN;
}

/* ***************************** Generics *********************************** */

Generic::Generic(const Runtime& r, const Atom& target, bool order, bool neg)
  : Function(r),
    redirTo(target), swapOrder(order), negate(neg)
{}

bool Generic::accepts(const Arguments& vals) const
{
  // We accept any two arguments
  return vals.size()==2;
}

Elob Generic::eval(const Arguments& vals) const
{
  // Process argument-swapping
  std::auto_ptr<Arguments> newValues(swapOrder ?
    new Arguments(vals.rbegin(), vals.rend()) :
    new Arguments(vals));

  // Find redirection-target and call
  const Function& f(rt.queryFunction(redirTo, *newValues));
  const Elob redirResult(f.eval(*newValues));

  // Check it is a bool and extract it
  if(redirResult.getType()!=rt.getBooleanType())
  {
    std::ostringstream msg;
    msg << "Comparison-function " << redirTo << " should return boolean value!";
    throw std::runtime_error(msg.str());
  }
  const bool boolResult(redirResult.getBoolean());

  // Return with inversion processed
  return Elob::forBoolean(negate ? !boolResult : boolResult, rt);
}

/* ********************************* Concrete ******************************* */

/* Number comparison */

bool NumberComp::accepts(const Arguments& vals) const
{
  // Two arguments
  if(vals.size()!=2)
    return false;

  // Both numbers
  Type::Ref numType(rt.getNumberType());
  if(!numType->isAssignableFrom(vals.front().getType()))
    return false;
  if(!numType->isAssignableFrom(vals.back().getType()))
    return false;

  return true;
}

Elob NumberComp::eval(const Arguments& vals) const
{
  assert(vals.size()==2);

  Type::Ref intType(rt.getIntegerType());
  Type::Ref numType(rt.getNumberType());

  // Both integers?
  if(intType->isAssignableFrom(vals.front().getType()) &&
     intType->isAssignableFrom(vals.back().getType()))
  {
    const Cell::intType a(vals.front().castTo(intType).getInteger());
    const Cell::intType b(vals.back().castTo(intType).getInteger());
    return Elob::forBoolean(equal ? a==b : a<b, rt);
  }

  // Both numbers
  assert(numType->isAssignableFrom(vals.front().getType()));
  assert(numType->isAssignableFrom(vals.back().getType()));
  const Cell::numberType a(vals.front().castTo(numType).getNumber());
  const Cell::numberType b(vals.back().castTo(numType).getNumber());
  return Elob::forBoolean(equal ? a==b : a<b, rt);
}

/* Callbacks */

Elob stringEqual(const Runtime& rt, const Arguments& vals)
{
  assert(vals.size()==2);
  const std::string& a(vals.front().getString());
  const std::string& b(vals.back().getString());
  return Elob::forBoolean(a==b, rt);
}

Elob stringLessThan(const Runtime& rt, const Arguments& vals)
{
  assert(vals.size()==2);
  const std::string& a(vals.front().getString());
  const std::string& b(vals.back().getString());
  return Elob::forBoolean(a<b, rt);
}

Elob boolEqual(const Runtime& rt, const Arguments& vals)
{
  assert(vals.size()==2);
  return Elob::forBoolean(vals[0].getBoolean()==vals[1].getBoolean(), rt);
}

} // namespace Comparisons

} // namespace ellib
} // namespace lispel
