/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef ELLIB_COMPARISONS_HPP
#define ELLIB_COMPARISONS_HPP

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "Function.hpp"

namespace lispel
{
namespace ellisp
{
  class Runtime;
  class Elob;
  class Atom;
}
using namespace ellisp;
namespace ellib
{

/**
 * Comparison-functions; only operator== and operator< must be defined for
 * custom types, the others are inferred by them automatically and in a generic
 * way.
 */
namespace Comparisons
{

  /**
   * Register those functions with a Runtime
   */
  void define(Runtime& rt);

  // Query for the names of == and <
  Atom getEqualName();
  Atom getLessThanName();

  /* *************************** Generics *********************************** */

  /**
   * Generic comparison, accepts every set of two arguments; which comparison
   * it is is determined by the parameters.
   */
  class Generic : public Function
  {

    public:

      /**
       * Create a new Generic comparison
       * @param r The runtime
       * @param target The name of the redirection-target function
       * @param order Whether to reorder the inputs when redirecting
       * @param neg Whether to negate the final result
       */
      Generic(const Runtime& r, const Atom& target, bool order, bool neg);

      virtual bool accepts(const Arguments& vals) const;
      virtual Elob eval(const Arguments& vals) const;

    private:

      /** The name of the operator to direct to */
      const Atom redirTo;

      /** Whether to swap the order of the inputs when redirecting */
      const bool swapOrder;

      /** Whether to negate the result finally */
      const bool negate;

  };

  /* ******************************* Concrete ******************************* */

  /**
   * Comparison-function for numbers, both equal and less-than for simplicity.
   */
  class NumberComp : public Function
  {
    public:
      inline NumberComp(const Runtime& rt, bool eq)
        : Function(rt), equal(eq)
      {}
      virtual bool accepts(const Arguments& vals) const;
      virtual Elob eval(const Arguments& vals) const;
    private:
      /** Compare for equal? (Less-than otherwise) */
      const bool equal;
  };

  // Callbacks for string-comparison and operator== for booleans
  Elob stringEqual(const Runtime& rt, const Arguments& vals);
  Elob stringLessThan(const Runtime& rt, const Arguments& vals);
  Elob boolEqual(const Runtime& rt, const Arguments& vals);

} // namespace Comparisons

} // namespace ellib
} // namespace lispel

#endif // Header-guard
