/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "Conditions.hpp"

#include "Atom.hpp"
#include "Elob.hpp"
#include "Runtime.hpp"

namespace lispel
{
namespace ellib
{

#define NAME_BOOLEAN Atom("Bedingung")
//#define NAME_IF Atom("Bedingte")
#define NAME_NOT Atom("Verneinung")
#define NAME_AND Atom("Konjunktion")
#define NAME_OR Atom("Disjunktion")
#define NAME_TRUE Atom("wahr")
#define NAME_FALSE Atom("falsch")

namespace Conditions
{

void define(Runtime& rt)
{
  // Boolean-type, if-function and true/false
  rt.addType(NAME_BOOLEAN, new Boolean(rt));
  //rt.addFunction(NAME_IF, new If(rt));
  rt.addConstant(NAME_TRUE, Elob::forBoolean(true, rt));
  rt.addConstant(NAME_FALSE, Elob::forBoolean(false, rt));

  /* Logic operators via callback-functions */

  // not(bool)
  Signature sig;
  sig.push_back(rt.getBooleanType());
  rt.addFunction(NAME_NOT, new CallbackFunction(rt, sig, &logicNot));

  // and/or(bool, bool)
  sig.clear();
  sig.push_back(rt.getBooleanType());
  sig.push_back(rt.getBooleanType());
  rt.addFunction(NAME_AND, new CallbackFunction(rt, sig, &logicAnd));
  rt.addFunction(NAME_OR, new CallbackFunction(rt, sig, &logicOr));
}

/* ****************************** Boolean *********************************** */

Boolean::Boolean(const Runtime& r)
  : Type(NAME_BOOLEAN, r)
{}

std::ostream& Boolean::print(std::ostream& out, const void* val) const
{
  return out << (val ? NAME_TRUE : NAME_FALSE);
}

#if 0
/* *********************************** If *********************************** */

bool If::accepts(const Arguments& vals) const
{
  // Need three arguments
  if(vals.size()!=3)
    return false;

  // First one must be condition
  return vals.front().getType()==getBooleanType(rt);
}

Elob If::eval(const Arguments& vals) const
{
  assert(vals.size()==3);
  return Boolean::getValue(vals[0]) ? vals[1] : vals[2];
}
#endif

/* ******************************** Callbacks ******************************* */

Elob logicNot(const Runtime& rt, const Arguments& args)
{
  assert(args.size()==1);
  return Elob::forBoolean(!args.front().getBoolean(), rt);
}

Elob logicAnd(const Runtime& rt, const Arguments& args)
{
  assert(args.size()==2);
  const bool val1(args[0].getBoolean());
  const bool val2(args[1].getBoolean());
  return Elob::forBoolean(val1 && val2, rt);
}

Elob logicOr(const Runtime& rt, const Arguments& args)
{
  assert(args.size()==2);
  const bool val1(args[0].getBoolean());
  const bool val2(args[1].getBoolean());
  return Elob::forBoolean(val1 || val2, rt);
}

} // namespace Conditions
} // namespace ellib

/* ********************************** Ellisp-code *************************** */

namespace ellisp
{

Type::Ref Runtime::getBooleanType() const
{
  return queryType(NAME_BOOLEAN);
}

/* For holding boolean values, the data pointer itself is abused; set to NULL
 * for false and something else (i.e., must be some valid "pointer") for true.
 */

Elob Elob::forBoolean(bool val, const Runtime& rt)
{
  void* dataVal(val ? &val : NULL); // Need "valid" address for true-value
  return Elob(rt.getBooleanType(), dataVal, rt);
}

bool Elob::getBoolean() const
{
  assert(getType()==getRuntime().getBooleanType());
  return static_cast<bool>(getValue());
}

} // namespace ellisp
} // namespace lispel
