/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008,2012  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "SideEffects.hpp"

#include "Atom.hpp"

#include <stdexcept>
#include <memory>
#include <typeinfo>

namespace lispel
{
namespace ellib
{

/**
 * Library-support for handling side-effect "values"
 */
namespace SideEffects
{

#define NAME_SIDEEFFECT_VALUE Atom("Side-Effect")
#define NAME_SEQUENCE Atom("Sequenz")
#define NAME_NOP Atom("nichts")
#define NAME_DO Atom("_do")

void define(Runtime& rt)
{
  rt.addType(NAME_SIDEEFFECT_VALUE, new SideEffectVal(rt));
  rt.addFunction(NAME_SEQUENCE, new CreateSeq(rt));

  // Define nop, abusing Sequence-class
  rt.addConstant(NAME_NOP, Elob(rt.getSideEffectType(), new Sequence(rt), rt));
  
  // do(side-effect)
  Signature sig;
  sig.push_back(rt.getSideEffectType());
  rt.addAction(NAME_DO, new CallbackAction(rt, sig, &doAction));
}


/* ************************** Side-effect value ***************************** */

SideEffectVal::SideEffectVal(const Runtime& r)
  : Type(NAME_SIDEEFFECT_VALUE, r)
{}

void SideEffectVal::destroyData(void* val) const
{
  delete static_cast<SideEffect*>(val);
}

std::ostream& SideEffectVal::print(std::ostream& out, const void* val) const
{
  const SideEffect& seVal(*static_cast<const SideEffect*>(val));

  // Only do special things for dynamic type LiteralAction
  try
  {
    const LiteralAction& laVal(dynamic_cast<const LiteralAction&>(seVal));
    return out << laVal;
  } catch(const std::bad_cast& exc)
  {}

  // Otherwise, use default implementation
  return Type::print(out, val);
}

/* ********************************* Sequence ******************************* */

void Sequence::perform() const
{
  for(actArr::const_iterator i(acts.begin()); i!=acts.end(); ++i)
  {
    // Values should be casted correctly
    assert(rt.getSideEffectType()==i->getType());
    const SideEffect* se(static_cast<const SideEffect*>(i->getValue()));
    se->perform();
  }
}

/* ********************************** CreateSeq ***************************** */

bool CreateSeq::accepts(const Arguments& vals) const
{
  // At least two arguments
  if(vals.size()<2)
    return false;

  // All of them must be side-effect values
  for(Arguments::const_iterator i(vals.begin()); i!=vals.end(); ++i)
    if(!rt.getSideEffectType()->isAssignableFrom(i->getType()))
      return false;

  return true;
}

Elob CreateSeq::eval(const Arguments& vals) const
{
  assert(vals.size()>=2);
  std::auto_ptr<Sequence> result(new Sequence(rt));
  for(Arguments::const_iterator i(vals.begin()); i!=vals.end(); ++i)
    result->add(i->castTo(rt.getSideEffectType()));

  return Elob(rt.getSideEffectType(), result.release(), rt);
}

/* ******************************** do-Action ******************************* */

void doAction(const Runtime& rt, Arguments& vals)
{
  assert(vals.size()==1);
  assert(rt.getSideEffectType()==vals[0].getType());

  static_cast<const SideEffect*>(vals[0].getValue())->perform();
}

} // namespace SideEffects
using namespace SideEffects;

} // namespace ellib
using namespace ellib;

/* *************************** Elob::forLiteralAction *********************** */

namespace ellisp
{

Elob Elob::forSideEffect(SideEffect* val, const Runtime& rt)
{
  return Elob(rt.getSideEffectType(), val, rt);
}

const SideEffect& Elob::getSideEffect() const
{
  assert(getType()==getRuntime().getSideEffectType());
  return *static_cast<const SideEffect*>(getValue());
}

Type::Ref Runtime::getSideEffectType() const
{
  return queryType(NAME_SIDEEFFECT_VALUE);
}

} // namespace ellisp
} // namespace lispel
