/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "Streams.hpp"
#include "Variables.hpp"

#include "Atom.hpp"
#include "Cell.hpp"

#include <fstream>
#include <sstream>
#include <stdexcept>
#include <vector>
#include <string>

namespace lispel
{
namespace ellib
{

namespace Streams
{

#define MAX_LINE 2048

#define NAME_STREAM Atom("Stream")
#define NAME_ISTREAM Atom("Inputstream")
#define NAME_OSTREAM Atom("Outputstream")

#define NAME_CREATE_BINARY_OFSTREAM Atom("Binaerdatei-Outputstream")
#define NAME_CREATE_TEXT_OFSTREAM Atom("Textdatei-Outputstream")
#define NAME_CREATE_BINARY_IFSTREAM Atom("Binaerdatei-Inputstream")
#define NAME_CREATE_TEXT_IFSTREAM Atom("Textdatei-Inputstream")
#define NAME_CREATE_ISSTREAM Atom("String-Inputstream")

#define NAME_PRINT Atom("gib-aus")
#define NAME_INPUT Atom("lese-ein")
#define NAME_WRITESTRING Atom("schreibe")
#define NAME_READLINE Atom("lese-Zeile")
#define NAME_READSTRINGBYLEN Atom("lese-String-nach-Laenge")
#define NAME_WRITEBYTE Atom("schreibe-Byte")
#define NAME_READBYTE Atom("lese-Byte")

#define NAME_EOF Atom("EOF")
#define NAME_STDOUT Atom("stdout")
#define NAME_STDERR Atom("stderr")
#define NAME_STDIN Atom("stdin")
#define NAME_NEWLINE Atom("newline")

void define(Runtime& rt)
{
  Signature sig;

  // Stream types
  rt.addType(NAME_STREAM, new Stream(rt));
  rt.addType(NAME_ISTREAM, new IOStream<std::istream>(NAME_ISTREAM, rt));
  rt.addType(NAME_OSTREAM, new IOStream<std::ostream>(NAME_OSTREAM, rt));

  // Creation
  rt.addFunction(NAME_CREATE_BINARY_OFSTREAM,  
    new CreateFileIOStream<std::ofstream>(rt, getOStreamType(rt),
      std::ios::binary | std::ios::trunc | std::ios::out));
  rt.addFunction(NAME_CREATE_TEXT_OFSTREAM,  
    new CreateFileIOStream<std::ofstream>(rt, getOStreamType(rt),
      std::ios::trunc | std::ios::out));
  rt.addFunction(NAME_CREATE_BINARY_IFSTREAM,  
    new CreateFileIOStream<std::ifstream>(rt, getIStreamType(rt),
      std::ios::binary | std::ios::in));
  rt.addFunction(NAME_CREATE_TEXT_IFSTREAM,  
    new CreateFileIOStream<std::ifstream>(rt, getIStreamType(rt),
      std::ios::in));
  
  // createStringIStream(string)
  sig.clear();
  sig.push_back(rt.getStringType());
  rt.addFunction(NAME_CREATE_ISSTREAM,
    new CallbackFunction(rt, sig, &createStringIStream));

  // Print / input
  rt.addAction(NAME_PRINT, new Print(rt));
  rt.addAction(NAME_INPUT, new Input(rt));

  // writeString(ostream, string)
  sig.clear();
  sig.push_back(getOStreamType(rt));
  sig.push_back(rt.getStringType());
  rt.addAction(NAME_WRITESTRING, new CallbackAction(rt, sig, &writeString));

  // readLine(istream, variable)
  sig.clear();
  sig.push_back(getIStreamType(rt));
  sig.push_back(Variables::getVariableType(rt));
  rt.addAction(NAME_READLINE, new CallbackAction(rt, sig, &readLine));

  // readStringByLen(istream, integer, variable)
  sig.clear();
  sig.push_back(getIStreamType(rt));
  sig.push_back(rt.getIntegerType());
  sig.push_back(Variables::getVariableType(rt));
  rt.addAction(NAME_READSTRINGBYLEN,
    new CallbackAction(rt, sig, &readStringByLen));

  // writeByte(ostream, integer)
  sig.clear();
  sig.push_back(getOStreamType(rt));
  sig.push_back(rt.getIntegerType());
  rt.addAction(NAME_WRITEBYTE, new CallbackAction(rt, sig, &writeByte));

  // readByte(istream, variable)
  sig.clear();
  sig.push_back(getIStreamType(rt));
  sig.push_back(Variables::getVariableType(rt));
  rt.addAction(NAME_READBYTE, new CallbackAction(rt, sig, &readByte));

  // Constants
  rt.addConstant(NAME_EOF, Elob::forInteger(-1, rt));
  rt.addConstant(NAME_NEWLINE, Elob::forString("\n", rt));
  rt.addConstant(NAME_STDOUT, Elob(getOStreamType(rt), &std::cout, rt, false));
  rt.addConstant(NAME_STDERR, Elob(getOStreamType(rt), &std::cerr, rt, false));
  rt.addConstant(NAME_STDIN, Elob(getIStreamType(rt), &std::cin, rt, false));
}

Type::Ref getStreamType(const Runtime& rt)
{
  return rt.queryType(NAME_STREAM);
}

Type::Ref getIStreamType(const Runtime& rt)
{
  return rt.queryType(NAME_ISTREAM);
}

Type::Ref getOStreamType(const Runtime& rt)
{
  return rt.queryType(NAME_OSTREAM);
}

/* ******************************** Types *********************************** */

Stream::Stream(const Runtime& r)
  : Type(NAME_STREAM, r)
{}

void Stream::destroyData(void* val) const
{
  // Can never happen, only sub-types are destroyed
  assert(false);
}

/* ******************************* Creation ********************************* */

Elob createStringIStream(const Runtime& rt, const Arguments& vals)
{
  // createStringIStream(string in)
  assert(vals.size()==1);
  assert(vals[0].getType()==rt.getStringType());
  return Elob(getIStreamType(rt),
    new std::istringstream(vals[0].getString()), rt);
}

/* ******************************** Print *********************************** */

bool Print::accepts(const Arguments& vals) const
{
  // print(ostream out, any val, ...)
  if(vals.size()<2)
    return false;
  if(!getOStreamType(rt)->isAssignableFrom(vals[0].getType()))
    return false;
  return true;
}

void Print::perform(Arguments& vals) const
{
  assert(vals.size()>=2);

  void* castStream(vals[0].castTo(getOStreamType(rt)).getValue());
  std::ostream& out(*static_cast<std::ostream*>(castStream));

  for(Arguments::size_type i(1); i!=vals.size(); ++i)
  {
    assert(i<vals.size());
    out << vals[i];
  }
}

/* ******************************** Input *********************************** */

bool Input::accepts(const Arguments& vals) const
{
  // print(istream out, variable+)

  // Check number arguments
  if(vals.size()<2)
    return false;

  // Check first one is input stream
  if(!getIStreamType(rt)->isAssignableFrom(vals[0].getType()))
    return false;

  // Check following ones are variables
  for(Arguments::size_type i(1); i!=vals.size(); ++i)
    if(!Variables::getVariableType(rt)->isAssignableFrom(vals[i].getType()))
      return false;

  return true;
}

void Input::perform(Arguments& vals) const
{
  assert(vals.size()>=2);

  void* castStream(vals[0].castTo(getIStreamType(rt)).getValue());
  std::istream& in(*static_cast<std::istream*>(castStream));

  for(Arguments::size_type i(1); i!=vals.size(); ++i)
  {
    assert(i<vals.size());
    Elob var(vals[i].castTo(Variables::getVariableType(rt)));

    Cell c(in);
    switch(c.getType())
    {
      case Cell::NUMBER:
        Variables::setVariable(var, Elob::forNumber(c.getNumber(), rt));
        break;
      case Cell::INTEGER:
        Variables::setVariable(var, Elob::forInteger(c.getInteger(), rt));
        break;
      case Cell::STRING:
        Variables::setVariable(var, Elob::forString(c.getString(), rt));
        break;
      default:
        throw std::runtime_error("Input only works for "
          " numbers, integers and strings!");
    }
  }
}

/* ******************************* IO-Actions ******************************* */

void writeString(const Runtime& rt, Arguments& vals)
{
  // writeString(ostream, string)
  assert(vals.size()==2);
  
  assert(vals[0].getType()==getOStreamType(rt));
  std::ostream& out(*static_cast<std::ostream*>(vals[0].getValue()));

  out << vals[1].getString();
}

void readLine(const Runtime& rt, Arguments& vals)
{
  // readLine(istream, variable)
  assert(vals.size()==2);

  assert(vals[0].getType()==getIStreamType(rt));
  std::istream& in(*static_cast<std::istream*>(vals[0].getValue()));

  char line[MAX_LINE];
  in.getline(line, MAX_LINE);

  Variables::setVariable(vals[1], Elob::forString(line, rt));
}

void readStringByLen(const Runtime& rt, Arguments& vals)
{
  // readStringByLen(istream, integer, variable) 
  assert(vals.size()==3);

  assert(vals[0].getType()==getIStreamType(rt));
  std::istream& in(*static_cast<std::istream*>(vals[0].getValue()));

  Cell::intType len(vals[1].getInteger());
  if(len<0)
  {
    std::ostringstream msg;
    msg << "Input-length can't be negative (" << len << ")!";
    throw std::runtime_error(msg.str());
  }

  std::vector<char> strData(len);
  in.read(&strData[0], len);

  std::string str(strData.begin(), strData.end());
  Variables::setVariable(vals[2], Elob::forString(str, rt));
}

void writeByte(const Runtime& rt, Arguments& vals)
{
  // writeByte(ostream, integer)
  assert(vals.size()==2);

  assert(vals[0].getType()==getOStreamType(rt));
  std::ostream& out(*static_cast<std::ostream*>(vals[0].getValue()));

  Cell::intType b(vals[1].getInteger());
  if(b<0 || b>0xFF)
  {
    std::ostringstream msg;
    msg << "Byte must be in the range [0, 255], but is " << b;
    throw std::runtime_error(msg.str());
  }

  out.put(b);
}

void readByte(const Runtime& rt, Arguments& vals)
{
  // readByte(istream, variable)
  assert(vals.size()==2);

  assert(vals[0].getType()==getIStreamType(rt));
  std::istream& in(*static_cast<std::istream*>(vals[0].getValue()));

  Variables::setVariable(vals[1], Elob::forInteger(in.get(), rt));
}

} // namespace Streams

} // namespace ellib
} // namespace lispel
