/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef ELLIB_STREAMS_HPP
#define ELLIB_STREAMS_HPP

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "Function.hpp"
#include "Action.hpp"
#include "Type.hpp"
#include "Runtime.hpp"
#include "Elob.hpp"

#include <iostream>

namespace lispel
{
namespace ellisp
{
  class Atom;
}
using namespace ellisp;
namespace ellib
{

/**
 * Input/output stream objects and actions to read/write/print.
 */
namespace Streams
{

  /**
   * Register those functions with a Runtime
   */
  void define(Runtime& rt);

  /**
   * Query for the stream types
   * @param rt The runtime
   * @return The input / output stream type of the given runtime
   */
  Type::Ref getStreamType(const Runtime& rt);
  Type::Ref getIStreamType(const Runtime& rt);
  Type::Ref getOStreamType(const Runtime& rt);

  /**
   * Type for basic stream, that is the common super type of istream and
   * ostream.
   */
  class Stream : public Type
  {
    public:
      Stream(const Runtime& r);
      virtual void destroyData(void* val) const;
      // No cast possible, we are at top of hierarchy
      // Printing is good in default way
  };

  /**
   * Type for input/output stream.
   * @param S the underlying stream-class to use
   */
  template<typename S>
    class IOStream : public Type
  {
    public:
      inline IOStream(const Atom& a, const Runtime& r)
        : Type(a, r)
      {
        addSuper(getStreamType(rt));
      }
      void destroyData(void* val) const
      {
        delete static_cast<S*>(val);
      }
      Elob castValueTo(const Elob& val, Type::Ref t) const
      {
        // Can only cast to stream
        assert(val.getType()==ref());
        assert(t==getStreamType(rt));

        S* ioval(static_cast<S*>(const_cast<void*>(val.getValue())));
        return Elob(t, static_cast<std::ios_base*>(ioval), val);
      }
      // Printing is good in default way
  };

  /**
   * Creating streams
   */
  template<typename S>
    class CreateFileIOStream : public SignatedFunction
  {
    public:
      inline CreateFileIOStream(const Runtime& r,
        Type::Ref t, std::ios::openmode m)
        : SignatedFunction(r), outType(t), mode(m)
      {
        arguments.push_back(rt.getStringType());
      }
    protected:
      virtual Elob evalCasted(const Arguments& vals) const
      {
        assert(vals.size()==1);
        assert(vals[0].getType()==rt.getStringType());
        return Elob(outType, new S(vals[0].getString().c_str(), mode), rt);
      }
    private:  
      const Type::Ref outType;
      const std::ios::openmode mode;
  };
  Elob createStringIStream(const Runtime& rt, const Arguments& vals);

  /**
   * Action to print (readable) to a stream a custom Elob.
   */
  class Print : public Action
  {
    public:
      explicit inline Print(const Runtime& r)
        : Action(r)
      {}
      virtual bool accepts(const Arguments& vals) const;
      virtual void perform(Arguments& vals) const;
  };

  /**
   * Action to input a basic Elob from a stream
   */
  class Input : public Action
  {
    public:
      explicit inline Input(const Runtime& r)
        : Action(r)
      {}
      virtual bool accepts(const Arguments& vals) const;
      virtual void perform(Arguments& vals) const;
  };

  /**
   * Basic IO actions:
   * - write string literally
   * - read line as string
   * - read string by length
   * - write single byte/char
   * - read single byte/char
   */
  void writeString(const Runtime& rt, Arguments& vals);
  void readLine(const Runtime& rt, Arguments& vals);
  void readStringByLen(const Runtime& rt, Arguments& vals);
  void writeByte(const Runtime& rt, Arguments& vals);
  void readByte(const Runtime& rt, Arguments& vals);

} // namespace Streams

} // namespace ellib
} // namespace lispel

#endif // Header-guard
