/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008,2012  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "Strings.hpp"

#include "Runtime.hpp"
#include "Elob.hpp"
#include "Atom.hpp"
#include "Cell.hpp"

#include <sstream>
#include <string>
#include <stdexcept>

namespace lispel
{
namespace ellib
{

namespace Strings
{

#define NAME_CAT Atom("Verkettung")
#define NAME_LENGTH Atom("Laenge")
#define NAME_SUBSTRING Atom("Substring")
#define NAME_CHARAT Atom("Indexzeichen")

void define(Runtime& rt)
{
  rt.addFunction(NAME_CAT, new Concatenation(rt));

  // length(string)
  Signature sig;
  sig.push_back(rt.getStringType());
  rt.addFunction(NAME_LENGTH, new CallbackFunction(rt, sig, &length));

  // substring(string, int, int)
  sig.clear();
  sig.push_back(rt.getStringType());
  sig.push_back(rt.getIntegerType());
  sig.push_back(rt.getIntegerType());
  rt.addFunction(NAME_SUBSTRING,
    new CallbackFunction(rt, sig, &substring));

  // charAt(string, int)
  sig.clear();
  sig.push_back(rt.getStringType());
  sig.push_back(rt.getIntegerType());
  rt.addFunction(NAME_CHARAT, new CallbackFunction(rt, sig, &charAt));
}

/* ******************************** Concatentation ************************** */

bool Concatenation::accepts(const Arguments& vals) const
{
  // We need at least one argument
  if(vals.empty())
    return false;

  // All arguments must be strings or integers (characters)
  for(Arguments::const_iterator i(vals.begin()); i!=vals.end(); ++i)
    if(!rt.getStringType()->isAssignableFrom(i->getType()) &&
       !rt.getIntegerType()->isAssignableFrom(i->getType()))
      return false;

  return true;
}

Elob Concatenation::eval(const Arguments& vals) const
{
  std::string result;
  for(Arguments::const_iterator i(vals.begin()); i!=vals.end(); ++i)
  {
    std::string cur;
    if(rt.getStringType()->isAssignableFrom(i->getType()))
      cur=i->castTo(rt.getStringType()).getString();
    else
      cur.push_back(static_cast<char>(
        i->castTo(rt.getIntegerType()).getInteger()));
    result.append(cur);
  }

  return Elob::forString(result, rt);
}

/* *********************************** Callbacks **************************** */

Elob length(const Runtime& rt, const Arguments& args)
{
  assert(args.size()==1);
  return Elob::forInteger(args.front().getString().length(), rt);
}

Elob substring(const Runtime& rt, const Arguments& args)
{
  assert(args.size()==3);
  const std::string& str(args[0].getString());
  const Cell::intType pos(args[1].getInteger());
  const Cell::intType len(args[2].getInteger());

  return Elob::forString(str.substr(pos, len), rt);
}

Elob charAt(const Runtime& rt, const Arguments& args)
{
  assert(args.size()==2);
  const std::string& str(args.front().getString());
  const Cell::intType ind(args.back().getInteger());

  if(ind<0 || ind>=str.length())
  {
    std::ostringstream msg;
    msg << "Index " << ind << " into string \"" << str << "\" is out of range!";
    throw std::out_of_range(msg.str());
  }

  return Elob::forInteger(static_cast<Cell::intType>(str[ind]), rt);
}

} // namespace Strings

} // namespace ellib
} // namespace lispel
