/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "Variables.hpp"

#include "Runtime.hpp"
#include "Elob.hpp"
#include "Atom.hpp"

namespace lispel
{
namespace ellib
{

namespace Variables
{

#define NAME_VARIABLE Atom("Variable")
#define NAME_VALUE Atom("Wert")
#define NAME_CREATEVAR Atom("Variable")
#define NAME_ASSIGNMENT Atom("setze")

void define(Runtime& rt)
{
  rt.addType(NAME_VARIABLE, new Variable(rt));
  rt.addAction(NAME_ASSIGNMENT, new Assignment(rt));
  rt.addFunction(NAME_CREATEVAR, new CreateVar(rt));

  // value(variable)
  Signature sig;
  sig.push_back(getVariableType(rt));
  rt.addFunction(NAME_VALUE, new CallbackFunction(rt, sig, &extractValue));
}

Type::Ref getVariableType(const Runtime& rt)
{
  return rt.queryType(NAME_VARIABLE);
}

/* ********************************** Variable ****************************** */

Variable::Variable(const Runtime& r)
  : Type(NAME_VARIABLE, r)
{}

void Variable::destroyData(void* val) const
{
  delete static_cast<Elob*>(val);
}

std::ostream& Variable::print(std::ostream& out, const void* val) const
{
  const Elob& eval(*static_cast<const Elob*>(val));
  out << "Variable holding " << eval;
  return out;
}

/* ********************************* CreateVar ****************************** */

bool CreateVar::accepts(const Arguments& vals) const
{
  // We want one argument, but its type doesn't matter
  return vals.size()==1;
}

Elob CreateVar::eval(const Arguments& vals) const
{
  assert(vals.size()==1);
  return Elob(getVariableType(rt), new Elob(vals[0]), rt);
}

/* ************************************ Value ******************************* */

Elob extractValue(const Runtime& rt, const Arguments& vals)
{
  assert(vals.size()==1);
  assert(vals.front().getType()==getVariableType(rt));
  return *static_cast<const Elob*>(vals.front().getValue());
}

/* ******************************* Assignment ******************************* */

bool Assignment::accepts(const Arguments& vals) const
{
  // We need two arguments
  if(vals.size()!=2)  
    return false;

  // First one should be variable
  if(vals[0].getType()!=getVariableType(rt))
    return false;

  // Second one does not matter
  return true;
}

void Assignment::perform(Arguments& vals) const
{
  assert(vals.size()==2);
  setVariable(vals[0], vals[1]);
}

void setVariable(Elob& var, const Elob& val)
{
  assert(var.getType()==getVariableType(var.getRuntime()));
  Elob& varVal(*static_cast<Elob*>(var.getValue()));
  varVal=val;
}

} // namespace Variables

} // namespace ellib
} // namespace lispel
