/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef ELLISP_ACTION_HPP
#define ELLISP_ACTION_HPP

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "Arguments.hpp"

namespace lispel
{
namespace ellisp
{

class Runtime;

/* ************************************************************************** */
/* ********************************* Action ********************************* */
/* ************************************************************************** */

/**
 * Actions are much like functions, except they convey side-effects and can not
 * have a return-value in turn.
 */
class Action
{

  /* ************************************************************************ */
  public:

    virtual ~Action();

    /**
     * Check if the action accepts a set of arguments
     * @param vals The arguments we want to hand in
     * @return True if they match the action's signature
     */
    virtual bool accepts(const Arguments& vals) const =0;

    /**
     * Perform the action with given arguments
     * @param vals The argument values, might be modified here!
     */
    virtual void perform(Arguments& vals) const =0;

  /* ************************************************************************ */
  protected:

    /** The Runtime we belong to */
    const Runtime& rt;

    /**
     * Construct an action for sub-classes
     * @param rt The runtime
     */
    explicit inline Action(const Runtime& r)
      : rt(r)
    {}

};

/* ************************************************************************** */
/* **************************** SignatedAction ****************************** */
/* ************************************************************************** */

/**
 * An action with fixed signature.
 * @see SignatedFunction
 */
class SignatedAction : public Action
{

  /* ************************************************************************ */
  public:

    /**
     * Checks simply with signature
     * @see Action::accepts(const Arguments&) const
     */
    virtual bool accepts(const Arguments& vals) const;

    /**
     * Does the casting and calls further with correctly casted types.
     * @see Action::perform(Arguments&) const
     */
    virtual void perform(Arguments& vals) const;

  /* ************************************************************************ */
  protected:

    /** The signature */
    Signature arguments;

    /**
     * Construct for sub-class with given signature
     * @param r The runtime
     * @param sig The signature
     */
    inline SignatedAction(const Runtime& r, const Signature& sig)
      : Action(r), arguments(sig)
    {}

    /**
     * Construct a signated action with (currently) empty signature, so it
     * can be constructed dynamically
     * @param r The runtime
     */
    explicit inline SignatedAction(const Runtime& r)
      : Action(r), arguments()
    {}

    /**
     * Perform in sub-class with casted arguments
     * @param vals The casted arguments conforming exactly to signature
     */
    virtual void performCasted(Arguments& vals) const =0;

};

/* ************************************************************************** */
/* **************************** CallbackAction ****************************** */
/* ************************************************************************** */

/**
 * Finally, a signated action backed by a simple function pointer.
 * @see CallbackFunction
 */
class CallbackAction : public SignatedAction
{

  /* ************************************************************************ */
  public:

    /**
     * Type of native callback
     * @param rt The runtime we are using
     * @param args The argument values
     */
    typedef void (*Callback)(const Runtime& rt, Arguments& args);

    /**
     * Construct it
     * @param r The runtime
     * @param sig The signature
     * @param cb The callback to use
     */
    inline CallbackAction(const Runtime& r, const Signature& sig, Callback cb)
      : SignatedAction(r, sig), native(cb)
    {}

  /* ************************************************************************ */
  protected:

    /**
     * Forward to callback
     * @see SignatedAction::performCasted(Arguments& args) const
     */
    virtual void performCasted(Arguments& args) const;

  /* ************************************************************************ */
  private:

    /** The callback registered */
    const Callback native;

};

} // namespace ellisp
} // namespace lispel

#endif // Header-guard
