/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "Arguments.hpp"
#include "List.hpp"
#include "Cell.hpp"
#include "Runtime.hpp"

#include <stdexcept>

namespace lispel
{
namespace ellisp
{

/* ************************************************************************** */
/* ******************************* Signature ******************************** */
/* ************************************************************************** */

bool Signature::matches(const Arguments& vals) const
{
  if(vals.size()!=size())
    return false;

  const_iterator sig(begin());
  Arguments::const_iterator val(vals.begin());
  for(; sig!=end(); ++sig, ++val)
  {
    assert(val!=vals.end());
    if((*sig)->isAnyType())
      continue;
    if(!(*sig)->isAssignableFrom(val->getType()))
      return false;
  }

  return true;
}

void Signature::upcast(const Arguments& vals, Arguments& dest) const
{
  // (Up-)cast to correct values for evaluation
  const_iterator sig(begin());
  Arguments::const_iterator val(vals.begin());
  for(; sig!=end(); ++sig, ++val)
  {
    assert(val!=vals.end());
    if(val->getType()==*sig || (*sig)->isAnyType())
      dest.push_back(*val);
    else 
      dest.push_back(val->castTo(*sig));
  }
}

void Signature::defineArgs(Runtime& rt,
  const ArgNames& names, const Arguments& vals) const
{
  assert(size()==vals.size());
  assert(vals.size()==names.size());

  ArgNames::const_iterator name(names.begin());
  Arguments::const_iterator val(vals.begin());
  for(; name!=names.end(); ++name, ++val)
  {
    assert(val!=vals.end());
    rt.addConstant(*name, *val);
  }
}

void Signature::parse(const Runtime& rt, const List* lst,
  Signature& sig, ArgNames& names)
{
  for(; lst; lst=lst->getCdr())
  {
    // Extract current cell as list
    if(lst->getCar().getType()!=Cell::LIST)
      throw std::runtime_error("Signature must consist of lists!");
    const List* curArg(lst->getCar().getList());

    // First part is type
    if(!curArg)
      throw std::runtime_error("In signature: Missing argument type!");
    if(curArg->getCar().getType()!=Cell::ATOM)
      throw std::runtime_error("In signature: Argument type must be Atom!");
    sig.push_back(rt.queryType(curArg->getCar().getAtom()));

    // Next one is argument name
    curArg=curArg->getCdr();
    if(!curArg)
      throw std::runtime_error("In signature: Missing argument name!");
    if(curArg->getCar().getType()!=Cell::ATOM)
      throw std::runtime_error("In signature: Argument name must be Atom!");
    names.push_back(curArg->getCar().getAtom());
  }
}

} // namespace ellisp
} // namespace lispel
