/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef ELLISP_ARGUMENTS_HPP
#define ELLISP_ARGUMENTS_HPP

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "Type.hpp"
#include "Elob.hpp"

#include <vector>

namespace lispel
{
namespace ellisp
{

class List;
class Runtime;

/* ************************************************************************** */
/* ******************************* Arguments ******************************** */
/* ************************************************************************** */

/**
 * Used to hold argument-values to function/action calls
 */
typedef std::vector<Elob> Arguments;

/**
 * Used for argument-names in a UserFunction/UserAction
 */
typedef std::vector<Atom> ArgNames;

/* ************************************************************************** */
/* ******************************* Signature ******************************** */
/* ************************************************************************** */

/**
 * A signature for either function or action; it is a set of argument types.
 * This class also collects the common code for SignatedFunction and
 * SignatedAction, like argument-upcasting and parsing of signature-lists.
 */
typedef std::vector<Type::Ref> signatureSup;
class Signature : public signatureSup
{

  /* ************************************************************************ */
  public:

    /**
     * Construct an empty signature
     */
    inline Signature()
      : signatureSup()
    {}

    // No destructor needed

    /**
     * Checks if this signature matches argument values
     * @see Function::accepts(const Arguments&) const
     */
    bool matches(const Arguments& vals) const;

    /**
     * Does the argument-upcasting so vals matches this signature exactly.
     * @param vals The values to cast
     * @param dest Store the casted values here
     */
    void upcast(const Arguments& vals, Arguments& dest) const;

    /**
     * Defines given arguments to their names in the scope given
     * @param rt The scope to define them in
     * @param names The argument names
     * @param vals The argument values
     */
    void defineArgs(Runtime& rt, 
      const ArgNames& names, const Arguments& vals) const;

    /**
     * Utility-function to parse a signature
     * @param rt The runtime to parse for
     * @param lst The signature to parse as List
     * @param sig Store the types here
     * @param names Store the names there
     */
    static void parse(const Runtime& rt, const List* lst,
      Signature& sig, ArgNames& names);

};

} // namespace ellisp
} // namespace lispel

#endif // Header-guard
