/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef ELLISP_ATOM_HPP
#define ELLISP_ATOM_HPP

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <string>
#include <map>
#include <iostream>

namespace lispel
{
namespace ellisp
{

/* ************************************************************************** */
/* ******************************* Atom ************************************* */
/* ************************************************************************** */

/**
 * An Atom is a identifier string (type, variable, class, whatever name)
 * mapped to an integer; this is to make comparison and passing of those
 * "identifiers" easier than using fully-fledged strings every time.
 *
 * Bookkeeping so that mapping int->string and string->int can be done fast is
 * internal.
 */
class Atom
{

  /* ************************************************************************ */
  public:

    /**
     * Create the Atom from a string.
     * @param str The string to map.
     */
    Atom(const std::string& str);

    /**
     * Default construct one, should be assigned to some other value to
     * initialize it.
     */
    inline Atom()
      : id(-1)
    {}

    // Copy an Atom by default copy-constructor
    // Assignment by default operator=

    /**
     * Map this Atom to its corresponding string.
     * @return The string to this Atom.
     */
    const std::string& str() const;

    /**
     * Compare two Atoms for equality.
     * @param a The first Atom.
     * @param b The second Atom.
     * @return True if they are equal.
     */
    friend inline bool operator==(const Atom& a, const Atom& b)
    {
      return a.id==b.id;
    }
    friend inline bool operator!=(const Atom& a, const Atom& b)
    {
      return !(a==b);
    }

    /**
     * Compare two Atoms; needed for using std::map with operator< as ordering.
     * @param a The first Atom
     * @param b The second Atom
     * @return True if the first is "less than" the second
     */
    friend inline bool operator<(const Atom& a, const Atom& b)
    {
      return a.id<b.id;
    }

    /**
     * Write it out to stream, simply as string
     * @param out The stream to write to
     * @param a The Atom to write
     * @return The stream
     */
    friend inline std::ostream& operator<<(std::ostream& out, const Atom& a)
    {
      return out << a.str();
    }

  /* ************************************************************************ */
  private:

    /** The internally used identifier of this Atom */
    int id;

    /* ****************** Bookkeeping of map ***************** */

    /** Next id to assign */
    static int nextId;

    /** Type of map for string->int */
    typedef std::map<std::string, int> idMapT;
    /** The map itself */
    static idMapT map;

};

} // namespace ellisp
} // namespace lispel

#endif // Header-guard
