/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008,2012  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "Cell.hpp"
#include "List.hpp"

#include <cctype>
#include <cstdlib>
#include <sstream>
#include <stdexcept>

namespace lispel
{
namespace ellisp
{

/* ************************************************************************** */
/* ******************************* Cell ************************************* */
/* ************************************************************************** */

Cell::~Cell()
{
  switch(typ)
  {
    // Free non-POD
    #define CELL_FREE_NONPOD(enumId, var) \
      case enumId: \
        assert(val.var); \
        delete val.var; \
        break;
    CELL_FREE_NONPOD(ATOM, atom)
    CELL_FREE_NONPOD(STRING, str)

    // List, might be NULL
    case LIST:
      if(val.lst)
        delete val.lst;
      break;

    // Nothing to delete
    case EMPTY:
    case INTEGER:
    case NUMBER:
      break;

    // Should not be reached
    default:
      assert(false);
  }
}

bool Cell::operator==(const Cell& o) const
{
  if(getType()!=o.getType())
    return false;

  switch(getType())
  {
    case ATOM:
      return *val.atom==*o.val.atom;
    case STRING:
      return *val.str==*o.val.str;
    case LIST:
      return List::equal(val.lst, o.val.lst);
    case NUMBER:
      return val.number==o.val.number;
    case INTEGER:
      return val.integer==o.val.integer;
    case EMPTY:
      return true;
#ifndef NDEBUG
    default:
      assert(false);
#endif
  }
}

#if 0
Should not be needed!
Cell& Cell::operator=(const Cell& c)
{
  // Eventually free current content 
  freeContent();

  // Assign new ones
  typ=c.typ;
  switch(typ)
  {
    // Create new non-POD
    case ATOM:
      val.atom=new Atom(c.getAtom());
      break;
    case STRING:
      val.str=new std::string(c.getString());
      break;
    
    // Plainly copy PODs
    case INTEGER:
      val.integer=c.getInteger();
      break;
    case NUMBER:
      val.number=c.getNumber();
      break;

    // Should not be reached!
    case EMPTY:
    default:
      assert(false);
  }

  return *this;
}
#endif // Disabled operator=

std::ostream& operator<<(std::ostream& out, const Cell& c)
{
  switch(c.typ)
  {
    // Integer/float just as-is
    case Cell::INTEGER:
      out << c.val.integer;
      break;
    case Cell::NUMBER:
      out << c.val.number;
      break;

    // Atom as corresponding string
    case Cell::ATOM:
      assert(c.val.atom);
      out << c.val.atom->str();
      break;

    // Lists have their own output routines
    case Cell::LIST:
      List::output(out, c.val.lst);
      break;

    // Strings enclosed in "" with quotation
    case Cell::STRING:
      assert(c.val.str);
      out << '\"';
      for(std::string::const_iterator i(c.val.str->begin());
        i!=c.val.str->end(); ++i)
      {
        switch(*i)
        {
          case '\\':
            out << "\\\\";
            break;
          case '\"':
            out << "\\\"";
            break;
          default:
            out << *i;
            break;
        }
      }
      out << '\"';
      break;

    // Should not be reached!
    case Cell::EMPTY:
    default:
      assert(false);
  }

  return out;
}

std::istream& operator>>(std::istream& in, Cell& c)
{
  assert(c.typ==Cell::EMPTY);

  // Skip whitespace
  while(std::isspace(in.get()));
  in.unget();

  // Check next character
  char next(in.peek());

  // List
  if(next=='(')
  {
    c.typ=Cell::LIST;
    c.val.lst=List::input(in);
  } else 
  
  // String
  if(next=='\"')
  {
    std::ostringstream str;
    in.get(); // Remove pending character, must be quote
    while((next=in.get())!='\"')
    {
      if(next=='\\')
        str.put(in.get());
      else
        str << next;
    }
    c.typ=Cell::STRING;
    c.val.str=new std::string(str.str());
  } else

  // Zahl
  if(std::isdigit(next) || next=='.' || next=='-')
  {
    bool first(true); // First character, that is, - allowed
    bool floating(false); // . found
    std::ostringstream str;
    while(next=in.get(),
      std::isdigit(next) ||
      (next=='.' && !floating) ||
      (next=='-' && first))
    {
      str << next;
      first=false;
      if(next=='.')
        floating=true;
    }
    in.unget();

    if(floating)
    {
      c.typ=Cell::NUMBER;
      c.val.number=std::atof(str.str().c_str());
    } else
    {
      c.typ=Cell::INTEGER;
      c.val.integer=std::atoi(str.str().c_str());
    }
  } else

  // Otherwise its an Atom
  #define IS_ATOM_CHAR(c) \
    (std::isalnum(c) || (c)=='_' || (c)=='-')
  if(IS_ATOM_CHAR(next) || next=='!') // '!' is atom, but no ! in the middle!
  {
    std::ostringstream str;
    str << next;
    in.get(); /* Get rid of the first character, unput() above */
    while(next=in.get(), IS_ATOM_CHAR(next))
      str << next;
    in.unget();

    c.typ=Cell::ATOM;
    c.val.atom=new Atom(str.str());
  }

  // Or it must be some error
  else
    throw std::runtime_error("Could not parse cell!");

  return in;
}

} // namespace ellisp
} // namespace lispel
