/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef ELLISP_CELL_HPP
#define ELLISP_CELL_HPP

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "Atom.hpp"

#include <string>
#include <iostream>
#include <cassert>

namespace lispel
{
namespace ellisp
{

class List;

/* ************************************************************************** */
/* ******************************* Cell ************************************* */
/* ************************************************************************** */

/**
 * Cell is a value used in the Lisp lists, that is, an Atom, a literal string
 * or a number (integer/floating point).
 */
class Cell
{

  /* ************************************************************************ */
  public:

    /** Possible types of a Cell */
    enum Type
    {
      ATOM, // Atom
      INTEGER, // Integer
      NUMBER, // Floating-point number
      STRING, // String
      LIST, // Sub-list
      EMPTY // No value yet, just for internal use
    };

    /** Type used to hold integer values */
    typedef int intType;

    /** Type used to hold real number values */
    typedef double numberType;

    // Constructor initializing with POD-type
    #define CELL_CONSTRUCTOR_POD(type, enumId, var) \
      explicit inline Cell(type v) \
        : typ(enumId) \
      { \
        val.var=v; \
      }
    // Constructor initializing with non-POD-type
    #define CELL_CONSTRUCTOR_NONPOD(clazz, enumId, var) \
      explicit inline Cell(const clazz& v) \
        : typ(enumId) \
      { \
        val.var=new clazz(v); \
      }
    // Define constructors
    CELL_CONSTRUCTOR_POD(intType, INTEGER, integer)
    CELL_CONSTRUCTOR_POD(numberType, NUMBER, number)
    CELL_CONSTRUCTOR_NONPOD(Atom, ATOM, atom)
    CELL_CONSTRUCTOR_NONPOD(std::string, STRING, str)

    // Constructor for List does not new but take the pointer directly
    explicit inline Cell(List* l)
      : typ(LIST)
    {
      val.lst=l;
    }

    /**
     * Construct an empty cell, mainly for input afterwards.
     */
    inline Cell()
      : typ(EMPTY)
    {}

    /** Read a Cell from an input stream
     * @param in The stream to read from
     */
    explicit inline Cell(std::istream& in)
      : typ(EMPTY)
    {
      in >> *this;
    }

    /**
     * Destroy, eventually freeing the new'ed non-POD contents.
     */
    ~Cell();

    /**
     * Compare a cell to an Atom.
     * @param c The Cell
     * @param a The Atom to compare to
     * @return True if c is an Atom and its value equals a.
     */
    friend inline bool operator==(const Cell& c, const Atom& a)
    {
      return c.typ==Cell::ATOM && a==c.getAtom();
    }

    /**
     * Return this cell's type
     * @return The cell's type
     */
    inline Type getType() const
    {
      assert(typ!=EMPTY); // Empty should only be used internally!
      return typ;
    }

    // Define accessors for POD-types
    #define CELL_ACCESSOR_POD(type, enumId, var, name) \
      inline type get##name() const \
      { \
        assert(typ==enumId); \
        return val.var; \
      }
    // Define accessors for non-POD-types
    #define CELL_ACCESSOR_NONPOD(clazz, enumId, var, name) \
      inline const clazz& get##name() const \
      { \
        assert(typ==enumId); \
        assert(val.var); \
        return *val.var; \
      }

    // Create accessors
    CELL_ACCESSOR_POD(intType, INTEGER, integer, Integer)
    CELL_ACCESSOR_POD(numberType, NUMBER, number, Number)
    CELL_ACCESSOR_NONPOD(Atom, ATOM, atom, Atom)
    CELL_ACCESSOR_NONPOD(std::string, STRING, str, String)

    // List-accessor may return NULL for empty list
    inline const List* getList() const
    {
      assert(typ==LIST);
      return val.lst;
    }
    inline List* getList()
    {
      assert(typ==LIST);
      return val.lst;
    }

    /**
     * Compare two cells
     * @param o The other cell
     * @return True if they are equal
     */
    bool operator==(const Cell& o) const;

    /**
     * Write the cell in classic Lisp notation out to a stream.
     * @param out The stream to write to
     * @param c The cell to output
     * @return The stream
     */
    friend std::ostream& operator<<(std::ostream& out, const Cell& c);

    /**
     * Read from the stream into the cell, it must be EMPTY!
     * @param in The stream to read from
     * @param c The cell to input
     * @return The stream
     */
    friend std::istream& operator>>(std::istream& in, Cell& c);

  /* ************************************************************************ */
  private:

    /** Type of this cell */
    Type typ;

    /** Value; non-POD types are instead pointed-to */
    union
    {
      Atom* atom;
      intType integer;
      numberType number;
      std::string* str;
      List* lst;
    } val;

    // No copying!
    inline Cell(const Cell& c)
    {}
    inline void operator=(const Cell& c)
    {}

};

} // namespace ellisp
} // namespace lispel

#endif // Header-guard
