/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef ELLISP_ELOB_HPP
#define ELLISP_ELOB_HPP

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "Type.hpp"
#include "Cell.hpp"

#include <cassert>
#include <string>
#include <iostream>

namespace lispel
{
namespace ellisp
{

class Runtime;
class SideEffect;

/* ************************************************************************** */
/* ******************************* Elob ************************************* */
/* ************************************************************************** */

/**
 * Elobs are the data-units used in Lispel's runtime-system; similar to the
 * Lisp-cells they can hold each of the built-in types and additionally they
 * might simply point to some user-defined object (from the native
 * runtime-library, like a vector-class, or user-created objects).  However,
 * they by themselves do not care about these values; they are simply
 * pointed-to data pieces correctly handled by the functions working with
 * them.
 *
 * User-created classes consist of an array of Elobs holding the super-objects
 * as well as all members; layout is determined and only needed for the
 * generated getters, the constructor and the free-method.
 *
 * Elobs use a reference-counting mechanism to determine lifetime and destruct
 * correctly; destruction of the value is done by a registered free-method.
 */
class Elob
{

  /* ************************************************************************ */
  public:

    /**
     * Construct a new Elob with reference-count 1.
     * @param t Type
     * @param val Data-value
     * @param rt The Runtime this belongs to (for the type)
     * @param destr Whether to destroy the value afterwards
     */
    Elob(Type::Ref t, void* val, const Runtime& rt, bool destr =true);

    /**
     * Construct a new Elob with a parent elob
     * @param t The type
     * @param val The data-value
     * @param p The parent
     */
    Elob(Type::Ref t, void* val, const Elob& p);

    /**
     * Construct an empty Elob "shared pointer" ready to assign a reference to
     */
    inline Elob()
      : fields(NULL)
    {}

    /**
     * Copy an existing Elob to get a new reference
     * @param e The elob to copy
     */
    inline Elob(const Elob& e)
      : fields(e.fields)
    {
      reference();
    }

    /**
     * Destructor: decrement reference count, freeing will be done there as
     * needed.
     */
    inline ~Elob()
    {
      release();
    }

    /**
     * Assignment is simply free current reference and reference the new object.
     * @param e The Elob to assign to this one
     * @return Reference to this
     */
    inline Elob& operator=(const Elob& e)
    {
      release(); // Ignores NULL fields
      fields=e.fields;
      reference();
      return *this;
    }

    /**
     * Retrieve the type of this Elob.
     * @return This object's type
     */
    inline Type::Ref getType() const
    {
      assert(fields);
      return fields->type;
    }

    /**
     * Retrieve this object's data value
     * @return The data value
     */
    inline const void* getValue() const
    {
      assert(fields);
      return fields->data;
    }
    inline void* getValue()
    {
      assert(fields);
      return fields->data;
    }

    /**
     * Retrieve the Runtime this belongs to
     * @return The runtime
     */
    inline const Runtime& getRuntime() const
    {
      assert(fields);
      assert(fields->rt);
      return *fields->rt;
    }

    /**
     * Cast this Elob to another type
     * @param t The target type
     * @return The casted Elob
     */
    Elob castTo(Type::Ref t) const;

    // Get values for built-in Cell types
    Cell::numberType getNumber() const;
    Cell::intType getInteger() const;
    const std::string& getString() const;
    const SideEffect& getSideEffect() const;
    bool getBoolean() const;

    /**
     * Print an Elob as text (for debugging/info)
     * @param out The stream to print to
     * @param e The Elob to print
     * @return The stream
     */
    friend std::ostream& operator<<(std::ostream& out, const Elob& e);

    // Create Elob's for built-in Cell types
    static Elob forNumber(Cell::numberType val, const Runtime& rt);
    static Elob forInteger(Cell::intType val, const Runtime& rt);
    static Elob forString(const std::string& val, const Runtime& rt);
    static Elob forSideEffect(SideEffect* val, const Runtime& rt);
    static Elob forBoolean(bool val, const Runtime& rt);

  /* ************************************************************************ */
  private:

    /**
     * This struct holds the contents; it is allocated once and shared between
     * copied instances with the same reference-count.
     */
    struct Content
    {

      /** The data-value stored herein */
      void* data;

      /** Type of the Elob */
      Type::Ref type;

      /** The runtime this value belongs to */
      const Runtime* rt;

      /** Destroy the value afterwards? */
      bool destroy;

      /** Parent Elob if any; release this instead of destroy value */
      Elob* parent;

      /** Reference count */
      int refcnt;

    };
    
    /** Pointer to contents */
    Content* fields;

    /**
     * Increment the reference counter
     */
    inline void reference()
    {
      // Might be called when copying empty Elob's
      if(fields)
        ++fields->refcnt;
    }

    /**
     * Decrement the reference counter
     */
    void release();

};

} // namespace ellisp
} // namespace lispel

#endif // Header-guard
