/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef ELLISP_FUNCTION_HPP
#define ELLISP_FUNCTION_HPP

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "Elob.hpp"
#include "Type.hpp"
#include "Arguments.hpp"

namespace lispel
{
namespace ellisp
{

class Runtime;

/* ************************************************************************** */
/* ****************************** Function ********************************** */
/* ************************************************************************** */

/**
 * Functions are at the heart of Lispel as a functional language.  Generally,
 * they have some set of argument-types they accept and can evaluate a set of
 * Elobs of those types to a return-Elob.
 */
class Function
{

  /* ************************************************************************ */
  public:

    virtual ~Function();

    /**
     * Check if this function accepts a given list of input-parameters (by their
     * types); this calls Type::isAssignableFrom, so in case of ambiguity an
     * exception might be thrown!
     * @param vals The arguments we want to hand in
     * @return True if they match the function's signature
     */
    virtual bool accepts(const Arguments& vals) const =0;

    /**
     * Apply this function to the given arguments.  The arguments must be
     * accepted as by accepts(const ArgValues&) const!
     * @param vals The argument values
     * @return The return value
     */
    virtual Elob eval(const Arguments& vals) const =0;

  /* ************************************************************************ */
  protected:

    /** The Runtime we belong to */
    const Runtime& rt;

    /**
     * Construct a function backed by native callback.
     * @param rt The runtime
     */
    explicit inline Function(const Runtime& r)
      : rt(r)
    {}

};

/* ************************************************************************** */
/* **************************** SignatedFunction **************************** */
/* ************************************************************************** */

/**
 * A function with a fixed signature; accepts and casting is done here, the
 * calculation itself is left over.
 */
class SignatedFunction : public Function
{

  /* ************************************************************************ */
  public:

    /**
     * Checks simply with signature
     * @see Function::accepts(const ArgValues&) const
     */
    virtual bool accepts(const Arguments& vals) const;

    /**
     * Does the casting and calls further with correctly casted types.
     * @see Function::eval(const ArgValues&) const
     */
    virtual Elob eval(const Arguments& vals) const;

  /* ************************************************************************ */
  protected:

    /** The signature */
    Signature arguments;

    /**
     * Construct for sub-class with given signature
     * @param r The runtime
     * @param sig The signature
     */
    inline SignatedFunction(const Runtime& r, const Signature& sig)
      : Function(r), arguments(sig)
    {}

    /**
     * Construct a signated function with (currently) empty signature, so it
     * can be constructed dynamically
     * @param r The runtime
     */
    explicit inline SignatedFunction(const Runtime& r)
      : Function(r), arguments()
    {}

    /**
     * Evaluate in sub-class with casted arguments
     * @param vals The casted arguments conforming exactly to signature
     * @return The calculated value
     */
    virtual Elob evalCasted(const Arguments& vals) const =0;

};

/* ************************************************************************** */
/* **************************** CallbackFunction **************************** */
/* ************************************************************************** */

/**
 * Finally, a signated function backed by a simple function pointer.
 */
class CallbackFunction : public SignatedFunction
{

  /* ************************************************************************ */
  public:

    /**
     * Type of native callback
     * @param rt The runtime we are using
     * @param args The argument values
     */
    typedef Elob (*Callback)(const Runtime& rt, const Arguments& args);

    /**
     * Construct it
     * @param r The runtime
     * @param sig The signature
     * @param cb The callback to use
     */
    inline CallbackFunction(const Runtime& r, const Signature& sig, Callback cb)
      : SignatedFunction(r, sig), native(cb)
    {}

  /* ************************************************************************ */
  protected:

    /**
     * Forward to callback
     * @see SignatedFunction::evalCasted(const ArgValues& args) const
     */
    virtual Elob evalCasted(const Arguments& args) const;

  /* ************************************************************************ */
  private:

    /** The callback registered */
    const Callback native;

};

} // namespace ellisp
} // namespace lispel

#endif // Header-guard
