/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "List.hpp"
#include "Cell.hpp"

#include <stdexcept>
#include <memory>
#include <cctype>
#include <cstdarg>

#include <sstream>

namespace lispel
{
namespace ellisp
{

List::~List()
{
  assert(car);
  delete car;

  if(cdr)
    delete cdr;
}

void List::pushBack(Cell* c)
{
  if(cdr)
    cdr->pushBack(c);
  else
    cdr=new List(c);
}

List* List::build(Cell* first, ...)
{
  // Start the list
  std::auto_ptr<List> result(new List(first));
  List* currentLast(result.get());

  // Append the optional arguments
  va_list args;
  va_start(args, first);
  Cell* nextCell;
  while(nextCell=va_arg(args, Cell*), nextCell)
  {
    currentLast->append(new List(nextCell));
    currentLast=currentLast->getCdr();
  }
  va_end(args);

  return result.release();
}

bool List::equal(const List* a, const List* b)
{
  // Both empty or both not!
  if(empty(a)!=empty(b))
    return false;

  // So both empty?
  if(empty(a))
    return true;

  // No, compare them car-by-car and cdr-by-cdr
  assert(!empty(b));
  assert(a && b);
  return a->getCar()==b->getCar() && equal(a->getCdr(), b->getCdr());
}

List* List::copy(const List* lst)
{
  // TODO: Real copy, without this hack!
  std::ostringstream out;
  output(out, lst);
  std::istringstream in(out.str());
  return input(in);
}

std::ostream& List::output(std::ostream& out, const List* l)
{
  out << '(';
  while(l)
  {
    out << l->getCar();
    l=l->getCdr();
    if(l)
      out << " ";
  }
  out << ')';
  return out;
}

List* List::input(std::istream& in)
{
  // Skip leading whitespace
  while(std::isspace(in.get()));
  in.unget();

  // Now, we should find an opening parenthesis
  if(in.get()!='(')
    throw std::runtime_error("Expected \'(\' at beginning of list!");

  // Start off with an empty list
  std::auto_ptr<List> result;
  List* curEnd(NULL);

  // Read in the cells until matching ) is found
  while(true)
  {
    while(std::isspace(in.get()));
    in.unget();
    if(in.get()==')')
      break;
    in.unget();

    List* newCdr(new List(new Cell(in)));
    if(curEnd)
      curEnd->append(newCdr);
    else
    {
      assert(!result.get());
      result.reset(newCdr);
    }
    curEnd=newCdr;
  }

  return result.release();
}

} // namespace ellisp
} // namespace lispel
