/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef ELLISP_LIST_HPP
#define ELLISP_LIST_HPP

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <iostream>
#include <cstddef>
#include <cassert>

namespace lispel
{
namespace ellisp
{

class Cell;

/* ************************************************************************** */
/* ******************************* List ************************************* */
/* ************************************************************************** */

/**
 * A LISP list, that is, linked list of Cells.  Empty list is represented by
 * NULL List*, should be checked by !
 */
class List
{

  /* ************************************************************************ */
  public:

    /**
     * Create a new list with given Cell value; this should be a new'ed pointer
     * which is taken over.  This way, we can re-use Cell's constructors for
     * List.  cdr will be set empty.
     * @param c The Cell-value
     * @param tail The list's tail
     */
    explicit inline List(Cell* c, List* tail =NULL)
      : car(c), cdr(tail)
    {
      assert(car);
    }

    /**
     * Deeply free this list.
     */
    ~List();

    /**
     * Retrieve the car-value.
     * @return The car-value
     */
    inline const Cell& getCar() const
    {
      return *car;
    }

    /**
     * Retrieve the cdr-value or NULL
     * @return The cdr-value, may be NULL
     */
    inline List* getCdr()
    {
      return cdr;
    }
    inline const List* getCdr() const
    {
      return cdr;
    }

    /**
     * Splice out the Cdr, returning the current one and setting it here to
     * NULL.
     * @return The old Cdr
     */
    inline List* spliceCdr()
    {
      List* result(cdr);
      cdr=NULL;
      return result;
    }

    /**
     * Append something to this list, that is, set cdr.
     * @param l The list to append
     */
    inline void append(List* l)
    {
      assert(l);
      assert(!cdr);
      cdr=l;
    }

    /**
     * Append a new cell to the very tail of this list
     * @param c The cell to append
     */
    void pushBack(Cell* c);

    /**
     * Convenience function for parser to build a List from a set of
     * Cell*'s.  Ownership is taken of the cells.
     * @param first The first cell to package up
     * @param ... The cells to package up to the list, ended by NULL
     * @return The List built therefrom
     */
    static inline List* build()
    {
      return NULL;
    }
    static List* build(Cell* first, ...);

    /**
     * Compare two lists.
     * @param a The first one
     * @param b The second one
     * @return True if they are equal
     */
    static bool equal(const List* a, const List* b);

    /**
     * Deep-copy a list
     * @param lst The list to copy, might be empty (NULL)
     * @return The copy
     */
    static List* copy(const List* lst);

    /**
     * Output a List.
     * @param out The output-stream to write to
     * @param l The list to output
     * @return The stream
     */
    static std::ostream& output(std::ostream& out, const List* l);

    /**
     * Input a List; pointer value should be NULL or holding nonesense, will
     * *not* be freed!
     * @param in The input-stream to read from
     * @param l The list-pointer to input to, might be set NULL for empty list!
     * @return The stream
     */
    static List* input(std::istream& in);

    /**
     * Check if a given list is empty.
     * @param l The list to check
     * @return True if it is empty
     */
    static inline bool empty(const List* l)
    {
      return !l;
    }

  /* ************************************************************************ */
  private:

    /** The car, this is the cell value */
    Cell* const car;

    /** The cdr, the following list */
    List* cdr;

};

} // namespace ellisp
} // namespace lispel

#endif // Header-guard
