/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008,2012  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "Runtime.hpp"
#include "Elob.hpp"
#include "List.hpp"
#include "Cell.hpp"
#include "Function.hpp"
#include "Cast.hpp"
#include "If.hpp"
#include "SideEffect.hpp"
#include "ConstDef.hpp"
#include "UserFunction.hpp"
#include "UserAction.hpp"
#include "UserClass.hpp"
#include "AnyType.hpp"

#include <memory>
#include <stdexcept>
#include <sstream>
#include <iostream>

namespace lispel
{
namespace ellisp
{

/* ************************************************************************** */
/* ****************************** Runtime *********************************** */
/* ************************************************************************** */

Runtime::Runtime()
  : parent(NULL)
{
  setOutputStream(std::cout);

  // We always need an AnyType
  (new AnyType(*this))->define(*this);
}

Runtime::Runtime(const Runtime& p)
  : parent(&p)
{
  setOutputStream(std::cout);
}

void Runtime::processAll(List* const lst)
{
  /* Function-/Action-/Class-Definition? */
  #define HANDLE_USER_DEF(type) \
    if(User##type::is##type(lst)) \
    { \
      std::auto_ptr<User##type> def(new User##type(*this, lst)); \
      def.release()->define(*this); /* This takes ownership */ \
      return; \
    }
  HANDLE_USER_DEF(Function)
  HANDLE_USER_DEF(Action)
  HANDLE_USER_DEF(Class)

  /* constant/action/evaluation */
  processActionBody(lst);
}

void Runtime::processActionBody(const List* lst)
{
  /* Constant-definition? */
  if(ConstDef::isConstDef(lst))
  {
    ConstDef def(*this, lst);
    def.perform();
    return;
  }

  /* Otherwise its some action/evaluation */
  processAction(lst);
}

void Runtime::processAction(const List* const lst) const
{
  /* Action-call? */
  #define EXCLUDE_SPECIAL(type) \
    && !type::is##type(lst)
  if(lst && lst->getCar().getType()==Cell::ATOM
    EXCLUDE_SPECIAL(Cast)
    EXCLUDE_SPECIAL(If)
    EXCLUDE_SPECIAL(LiteralAction)
    )
  {
    // First element is name
    const Atom& name(lst->getCar().getAtom());

    // Next the argument values
    Arguments args;
    for(const List* i(lst->getCdr()); i; i=i->getCdr())
      args.push_back(evaluate(i->getCar()));

    // Try to find this action and call it
    try
    {
      queryAction(name, args).perform(args);
      return;
    } catch(const Unknown& u)
    {
      /*
      std::cerr << "While evaluating as action:" << std::endl;
      std::cerr << u.what() << std::endl;
      std::cerr << "Trying as function." << std::endl;
      */
    }
  }

  /* Finally, try as evaluation and print value */
  Elob val(evaluate(lst));
  printResult(val);
}

void Runtime::process(Cell& c)
{
  // Process as list
  if(c.getType()==Cell::LIST)
  {
    processAll(c.getList());
    return;
  }

  // Evaluate and print otherwise
  Elob val(evaluate(c));
  printResult(val);
}

Elob Runtime::evaluate(const List* lst) const
{
  // Handle casts and ifs
  #define HANDLE_SPECIAL_VALUE(name) \
    if(name::is##name(lst)) \
    { \
      name val(*this, lst); \
      return val.evaluate(); \
    }
  HANDLE_SPECIAL_VALUE(Cast)
  HANDLE_SPECIAL_VALUE(If)

  // Action-literals / sequences
  if(LiteralAction::isLiteralAction(lst))
    return Elob::forSideEffect(new LiteralAction(*this, lst), *this);
  if(Sequence::isSequence(lst))
    return Elob::forSideEffect(new Sequence(*this, lst), *this);

  // Empty list can not be evaluated
  if(!lst)
    throw std::runtime_error("Empty list is no function call!");

  // First element is function to call
  if(lst->getCar().getType()!=Cell::ATOM)
    throw std::runtime_error("First element of function call must be atom!");
  const Atom& name(lst->getCar().getAtom());

  // Next elements form arguments
  Arguments args;
  while(lst=lst->getCdr(), lst)
    args.push_back(evaluate(lst->getCar()));

  // Find and call function
  return queryFunction(name, args).eval(args);
}

Elob Runtime::evaluate(const Cell& c) const
{
  switch(c.getType())
  {
    case Cell::LIST:
      return evaluate(c.getList());
    case Cell::ATOM:
      return queryConstant(c.getAtom());
    case Cell::NUMBER:
      return Elob::forNumber(c.getNumber(), *this);
    case Cell::INTEGER:
      return Elob::forInteger(c.getInteger(), *this);
    case Cell::STRING:
      return Elob::forString(c.getString(), *this);
    default:
      break;
  }
  assert(false); // Catches "default"
}

const Elob& Runtime::queryConstant(const Atom& name) const
{
  // Check this one first
  try
  {
    return consts.query(name);
  } catch(const Unknown& u)
  {
    // Not found, try parent
    if(!parent)
      throw;
    return parent->queryConstant(name);
  }
}

Type::Ref Runtime::queryType(const Atom& name) const
{
  // See queryConstant
  try
  {
    return types.query(name);
  } catch(const Unknown& u)
  {
    if(!parent)
      throw;
    return parent->queryType(name);
  }
}

const Function& Runtime::queryFunction(const Atom& name,
  const Arguments& vals) const
{
  // See queryConstant
  try
  {
    return funcs.query(name, vals);
  } catch(const Unknown& u)
  {
    if(!parent) 
      throw;
    return parent->queryFunction(name, vals);
  }
}

const Action& Runtime::queryAction(const Atom& name,
  const Arguments& vals) const
{
  // See queryConstant
  try
  {
    return actions.query(name, vals);
  } catch(const Unknown& u)
  {
    if(!parent) 
      throw;
    return parent->queryAction(name, vals);
  }
}

// getXXXType is implemented in runtime-library

} // namespace ellisp
} // namespace lispel
