/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef ELLISP_RUNTIME_HPP
#define ELLISP_RUNTIME_HPP

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "SignatedTable.hpp"
#include "TypeTable.hpp"
#include "ConstTable.hpp"
#include "Arguments.hpp"
#include "Function.hpp"
#include "Action.hpp"

#include <stdexcept>
#include <string>
#include <iostream>

namespace lispel
{
namespace ellisp
{

class Elob;
class List;
class Cell;
class Atom;

/* ************************************************************************** */
/* ****************************** Runtime *********************************** */
/* ************************************************************************** */

/**
 * The runtime environment holding the function table, variables currently
 * defined/in scope and other dynamically defined things.  Is also the main
 * Lisp processor.
 */
class Runtime
{

  /* ************************************************************************ */
  public:

    /**
     * Construct an "empty" runtime environment.
     */
    Runtime();

    /**
     * Construct a runtime with a given "parent"; if some entry is not defined
     * here, the parent is queried for it.  This way a stack of scopes can be
     * built.
     * @param p The parent runtime
     */
    explicit Runtime(const Runtime& p);

    // No destructor needed
    
    /**
     * Add a new object to one of the tables
     * @param name The object's Map-name
     * @param f/t/val The object to add
     */
    inline void addFunction(const Atom& name, Function* f)
    { 
      funcs.add(name, f);
    }
    inline void addAction(const Atom& name, Action* a)
    {
      actions.add(name, a);
    }
    inline void addType(const Atom& name, Type* t)
    {
      types.add(name, t);
    }
    inline void addConstant(const Atom& name, const Elob& val)
    {
      consts.add(name, val);
    }

    /**
     * Query for an object from a table, respecting eventual parent-Runtime
     * in case it is not defined here.
     * @param name The object's name
     * [@param vals] For functions, the argument-values
     * @return The found object
     * @throws Unknown if it is not defined
     */
    Type::Ref queryType(const Atom& name) const;
    const Function& queryFunction(const Atom& name,
      const Arguments& vals) const;
    const Action& queryAction(const Atom& name, const Arguments& vals) const;
    const Elob& queryConstant(const Atom& name) const;

    /**
     * Process a List of input according to its "type".  Depending on the
     * version, only actions described are performed or also definitions of
     * functions/actions/constants and the like.
     * @param lst The List to process
     * @throws std::runtime_error in case of error
     */
    void processAll(List* lst);
    void processActionBody(const List* lst);
    void processAction(const List* lst) const;

    /**
     * Process a Cell; for lists, it is processAll, for others it interprets
     * them as values and prints them out.
     * @param c The cell
     * @throws std::runtime_error in case of error
     */
    void process(Cell& c);
    
    /**
     * Evaluate a List as function call.
     * @param lst The List to evaluate
     * @return The created Elob value of this List
     * @throws std::runtime_error in case of error
     */
    Elob evaluate(const List* lst) const;

    /**
     * Evaluate a Cell, switching between its possible types.
     * @param c The Cell to evaluate
     * @return The created Elob value of the Cell
     * @throws std::runtime_error in case of error
     */
    Elob evaluate(const Cell& c) const;

    /**
     * Set the output stream
     * @param o The new output stream
     */
    inline void setOutputStream(std::ostream& o)
    {
      out=&o;
    }

    // Query for cell-types
    Type::Ref getIntegerType() const;
    Type::Ref getNumberType() const;
    Type::Ref getStringType() const;
    Type::Ref getSideEffectType() const;
    Type::Ref getBooleanType() const;

    /**
     * Exception thrown when something "named" in a table is not defined
     */
    class Unknown : public std::runtime_error
    {
      public:
        explicit inline Unknown(const std::string& what)
          : std::runtime_error("Not defined: "+what)
        {}
    };

  /* ************************************************************************ */
  private:

    /** The function table */
    SignatedTable<Function, Unknown> funcs;

    /** The action table */
    SignatedTable<Action, Unknown> actions;

    /** The defined types */
    TypeTable types;

    /** The defined constants */
    ConstTable consts;

    /** Our parent or NULL if no such */
    const Runtime* parent;

    /** The stream to use for value-outputting, not owned */
    std::ostream* out;

    /**
     * Print a result helper function
     * @param val The value to print
     */
    inline void printResult(const Elob& val) const
    {
      *out << val << std::endl;
    }

};

} // namespace ellisp
} // namespace lispel

#endif // Header-guard
