/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "SideEffect.hpp"
#include "Runtime.hpp"
#include "List.hpp"
#include "Cell.hpp"

#include <stdexcept>
#include <sstream>

namespace lispel
{
namespace ellisp
{

#define NAME_LITACT Atom("!")
#define NAME_SEQUENCE Atom("_sequence")

/* ************************************************************************** */
/* ***************************** SideEffect ********************************* */
/* ************************************************************************** */

bool SideEffect::accepts(const Arguments& vals) const
{
  return vals.empty();
}

void SideEffect::perform(Arguments& vals) const
{
  assert(vals.empty());
  perform();
}

/* ************************************************************************** */
/* **************************** LiteralAction ******************************* */
/* ************************************************************************** */

LiteralAction::LiteralAction(const Runtime& r, const List* lst)
  : SideEffect(r), args()
{
  assert(isLiteralAction(lst));
  assert(lst);

  // Next is name to call
  lst=lst->getCdr();
  if(!lst)
    throw std::runtime_error("Missing name to call for literal action!");
  if(lst->getCar().getType()!=Cell::ATOM)
    throw std::runtime_error("No Atom as name to call in literal action!");
  name=lst->getCar().getAtom();

  // Then comes the arguments
  while(lst=lst->getCdr(), lst)
    args.push_back(rt.evaluate(lst->getCar()));
}

void LiteralAction::perform() const
{
  // Find the action to call
  const Action& act(rt.queryAction(name, args));

  // Call it
  act.perform(args);
}

bool LiteralAction::isLiteralAction(const List* lst)
{
  // Must be (! ...)
  if(!lst)
    return false;
  if(lst->getCar().getType()!=Cell::ATOM)
    return false;
  if(lst->getCar().getAtom()!=NAME_LITACT)
    return false;

  return true;
}

std::ostream& operator<<(std::ostream& out, const LiteralAction& la)
{
  out << "(! " << la.name;
  for(Arguments::const_iterator i(la.args.begin());
    i!=la.args.end(); ++i)
  {
    out << " " << *i;
  }
  out << ")";

  return out;
}

/* ************************************************************************** */
/* ******************************* Sequence ********************************* */
/* ************************************************************************** */

Sequence::Sequence(const Runtime& r, const List* lst)
  : SideEffect(r)
#ifndef NDEBUG
  , actions(NULL)
#endif
{
  assert(isSequence(lst));
  actions=List::copy(lst->getCdr());
}

Sequence::~Sequence()
{
  if(actions)
    delete actions;
}

void Sequence::perform() const
{
  for(const List* i(actions); i; i=i->getCdr())
  {
    const Elob val(rt.evaluate(i->getCar()));
    if(!rt.getSideEffectType()->isAssignableFrom(val.getType()))
    {
      std::ostringstream msg;
      msg << "Expected side-effect in sequence, got '";
      msg << val << "'!";
      throw std::runtime_error(msg.str());
    }

    val.castTo(rt.getSideEffectType()).getSideEffect().perform();
  }
}

//std::ostream& operator<<(std::ostream& out, const Sequence& seq);

bool Sequence::isSequence(const List* lst)
{
  if(!lst)
    return false;
  if(lst->getCar().getType()!=Cell::ATOM)
    return false;
  if(lst->getCar().getAtom()!=NAME_SEQUENCE)
    return false;

  return true;
}

} // namespace ellisp
} // namespace lispel
