/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef ELLISP_SIDEEFFECT_HPP
#define ELLISP_SIDEEFFECT_HPP

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "Action.hpp"
#include "Arguments.hpp"
#include "Atom.hpp"

#include <iostream>

namespace lispel
{
namespace ellisp
{

class Runtime;
class List;

/* ************************************************************************** */
/* ***************************** SideEffect ********************************* */
/* ************************************************************************** */

/**
 * A side-effect, as the result of an Action, much like the return value of a
 * function; it is in fact an Action without arguments.  Used to implement
 * action-side-effects as values with support-functions by the library and for
 * "action-literals" like (! print stdout "Hello, world")
 */
class SideEffect : public Action
{

  /* ************************************************************************ */
  public:

    /**
     * Accepts empty arguments
     * @see Action::accepts(const Arguments&) const
     */
    virtual bool accepts(const Arguments& vals) const;

    /**
     * Forward call to empty perform()
     * @see Action::perform(Arguments&) const
     */
    virtual void perform(Arguments& vals) const;

    /**
     * Perform it; this is what sub-classes must override.
     */
    virtual void perform() const =0;

  /* ************************************************************************ */
  protected:

    /**
     * Construct for sub-classes
     * @param r The runtime
     */
    explicit inline SideEffect(const Runtime& r)
      : Action(r)
    {}

};

/* ************************************************************************** */
/* **************************** LiteralAction ******************************* */
/* ************************************************************************** */

/**
 * A literal action, that is, an action call encapsulated into a value.
 */
class LiteralAction : public SideEffect
{

  /* ************************************************************************ */
  public:

    /**
     * Construct a LiteralAction from its List-representation
     * @param r The runtime
     * @param lst The List to parse for it
     */
    LiteralAction(const Runtime& r, const List* lst);

    /**
     * @see SideEffect::perform() const
     */
    virtual void perform() const;

    /**
     * Print it out friendly
     * @param out The stream
     * @param la The literal action to print
     * @return The stream
     */
    friend std::ostream& operator<<(std::ostream& out, const LiteralAction& la);

    /**
     * Check if a List represents a LiteralAction
     * @param lst The List to check
     * @return True if it represents a LiteralAction
     */
    static bool isLiteralAction(const List* lst);

  /* ************************************************************************ */
  private:

    /** The name of the action to call */
    Atom name;

    /** The arguments to substitute to the action */
    mutable Arguments args;

};

/* ************************************************************************** */
/* ******************************* Sequence ********************************* */
/* ************************************************************************** */

/**
 * A sequence of side-effects combined into one, it does lazy evaluation like
 * _if, therefore it needs special care.
 */
class Sequence : public SideEffect
{

  /* ************************************************************************ */
  public:

    /**
     * Construct a Sequence from a List
     * @param r The runtime
     * @param lst The List to parse for it
     */
    Sequence(const Runtime& r, const List* lst);

    /**
     * Destroy, need to free the actions-list
     */
    ~Sequence();

    /**
     * @see SideEffect::perform() const
     */
    virtual void perform() const;

    /**
     * Print it out friendly
     * @param out The stream
     * @param seq The sequence to print
     * @return The stream
     */
    //friend std::ostream& operator<<(std::ostream& out, const Sequence& seq);

    /**
     * Check if a List represents a Sequence
     * @param lst The List to check
     * @return True if it represents a Sequence
     */
    static bool isSequence(const List* lst);

  /* ************************************************************************ */
  private:

    /** The list of values to evaluate and execute in sequence */
    List* actions;

};

} // namespace ellisp
} // namespace lispel

#endif // Header-guard
