/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef ELLISP_SIGNATEDTABLE_HPP
#define ELLISP_SIGNATEDTABLE_HPP

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "Atom.hpp"
#include "Arguments.hpp"

#include <map>
#include <vector>
#include <sstream>
#include <stdexcept>

namespace lispel
{
namespace ellisp
{

/* ************************************************************************** */
/* ************************** SignatedTable ********************************* */
/* ************************************************************************** */

/**
 * A table of signated "objects" (i.e., functions and actions):  This is a
 * two-step container, first a map mapping names to lists of overloaded
 * objects, next a list of equally-named objects only differentiated by their
 * signature.
 * @param T type of objects stored
 * @param Unknown exception to throw when queried for unknown object
 */
template<typename T, typename Unknown>
  class SignatedTable
{

  /* ************************************************************************ */
  public:

    /**
     * Create an empty table
     */
    inline SignatedTable()
      : table()
    {}

    /**
     * Destruct the table
     */
    ~SignatedTable()
    {
      for(typename signMap::iterator i(table.begin()); i!=table.end(); ++i)
        for(typename signArr::iterator j(i->second.begin());
          j!=i->second.end(); ++j)
        {
          delete *j;
        }
    }

    /**
     * Add a new object to the table
     * @param name The name of it
     * @param f The object; it is taken ownership of
     */
    inline void add(const Atom& name, T* f)
    {
      table[name].push_back(f);
    }

    /**
     * Query an object with a given name accepting a given argument list.
     * @param name The name
     * @param args The argument list
     * @return The corresopnding object
     * @throws Runtime::Unknown if no such object is found
     * @throws std::runtime_error if this is ambiguous
     */
    const T& query(const Atom& name, const Arguments& args) const
    {
      typename signMap::const_iterator pos(table.find(name));
      if(pos==table.end()) // Nothing with this name found
        throw Unknown("Function or action named "+name.str());

      T* result(NULL); // Store the result here when found
      for(typename signArr::const_iterator i(pos->second.begin());
        i!=pos->second.end(); ++i)
      {
        if((*i)->accepts(args))
        {
          // Ambiguity!
          if(result)
          {
            std::ostringstream msg;
            msg << "Multiple functions/actions '" << name.str();
            msg << "' match arguments!";
            throw std::runtime_error(msg.str());
          }

          result=*i;
        }
      }

      // Nothing suitable found
      if(!result)
      {
        std::ostringstream msg;
        msg << "Function or action named " << name << " matching arguments";
        throw Unknown(msg.str());
      }

      return *result;
    }

  /* ************************************************************************ */
  private:

    /** Lists overloaded objects with same name */
    typedef std::vector<T*> signArr;

    /** The mapping name -> list of overloaded objects */
    typedef std::map<Atom, signArr> signMap;

    /** The map used herein */
    signMap table;

};

} // namespace ellisp
} // namespace lispel

#endif // Header-guard
