/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "Type.hpp"
#include "AnyType.hpp"
#include "Elob.hpp"

#include <stdexcept>
#include <sstream>

namespace lispel
{
namespace ellisp
{

/* ************************************************************************** */
/* ******************************* Type ************************************* */
/* ************************************************************************** */

Type::~Type()
{
  // Nothing to be done here, super-type-array only contains not owned pointers
}

void Type::addSuper(Ref sup)
{
  if(isAssignableFrom(sup))
    throw std::invalid_argument("Type can't be assignable from super-type!");
  supers.push_back(sup);
}
    
bool Type::isAssignableFrom(Ref sub) const
{
  // Type is assignable from itself
  /* this can't trigger ambiguity because it can't then be assignable from any
   * super type (because they are it's own)
   */
  if(ref()==sub)
    return true;

  // Otherwise, check the sub's supers' assignability
  bool superFound(false); // We want exactly one!
  for(supersT::const_iterator i(sub->supers.begin()); i!=sub->supers.end(); ++i)
    if(isAssignableFrom(*i))
    {
      // Ambiguity!
      if(superFound)
      {
        std::ostringstream msg;
        msg << "Assignability of " << name << " from " << sub->name;
        msg << " is ambiguous!" << std::endl;
        throw std::runtime_error(msg.str());
      }

      // Now we've found one
      superFound=true;
    }

  return superFound;
}

void Type::destroyData(void* val) const
{
  // Do nothing by default
}

Elob Type::castValueTo(const Elob& val, Ref t) const
{
  assert(val.getType()==ref());

  std::ostringstream msg;
  msg << "Attempted to cast " << val << " to " << t->name.str();
  msg << " via Type::castValueTo!";
  throw std::runtime_error(msg.str());
}

std::ostream& Type::print(std::ostream& out, const void* val) const
{
  return out << "(" << name << ")" << val;
}

bool Type::isAnyType() const
{
  return ref()==AnyType::getAnyType(rt);
}

} // namespace ellisp
} // namespace lispel
