/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef ELLISP_TYPE_HPP
#define ELLISP_TYPE_HPP

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "Atom.hpp"

#include <vector>
#include <iostream>

namespace lispel
{
namespace ellisp
{

class Elob;
class Runtime;

/* ************************************************************************** */
/* ******************************* Type ************************************* */
/* ************************************************************************** */

/**
 * A type as used in Lispel; types build a hierarchy much like the C++ class
 * system; possibly a forest of seperate trees, possibly multiple inheritance
 * in both ways.
 *
 * During runtime, there should be exactly one instance of this class for each
 * defined type, and they should be passed/stored/compared simply by their
 * pointers.
 */
class Type
{

  /* ************************************************************************ */
  public:

    /** Use this to refer to, store and compare types; hide the pointer */
    typedef const Type* Ref;

    /**
     * Construct a new Type for the given Atom.  By now, it has no super-types
     * and must be given ones.
     * @param a This type's name
     * @param r The runtime this type belongs to
     */
    inline Type(const Atom& a, const Runtime& r)
      : rt(r), name(a), supers()
    {}

    /**
     * Destructor so it can be virtual
     */
    virtual ~Type();
    
    /**
     * Retrieve the type-reference for a type
     * @return The reference for this type
     */
    inline Ref ref() const
    {
      return this;
    }

    /**
     * Query for the type's name, for output
     * @return The type's name
     */
    inline const Atom& getName() const
    {
      return name;
    }

    /**
     * Add a super-type.  This type must not be assignable from the super-type
     * (i.e., it must not appear in its own super-hierarchy), otherwise an
     * exception is thrown.
     * @param sup The new super type
     * @throws std::invalid_argument if illegal
     */
    void addSuper(Ref sup);
    
    /**
     * Check if this type is assignable from some other type (i.e., the other
     * type has this one somewhere in its super-hierarchy).  If ambiguous (this
     * type occurs multiple times in the other's super types) an exception is
     * thrown.
     * @param sub The other type to assign from to this type
     * @return True if this is unambiguous assignable from sub
     * @throws std::runtime_error when assignment would be ambiguous.
     */
    bool isAssignableFrom(Ref sub) const;

    /**
     * Destroy a Elob-data-pointer of a thing of this type; does nothing by
     * default.
     * @param val The data-value
     */
    virtual void destroyData(void* val) const;

    /**
     * Cast an Elob data value of this type to a new type; the new type must be
     * assignable from this one, otherwise an exception might be thrown.  The
     * default implementation always throws.
     * @param val The data value as Elob
     * @param t The target-type
     * @return An Elob of val casted to t.
     * @throws std::runtime_error if this cast is not possible.
     */
    virtual Elob castValueTo(const Elob& val, Ref t) const;

    /**
     * Print out a value to a stream for debugging/info; per default simply the
     * pointer and the type name.
     * @param out The stream
     * @param val The data value
     * @return The stream
     */
    virtual std::ostream& print(std::ostream& out, const void* val) const;

    /**
     * Check if this type is the anytype of the runtime
     * @return True if it is.
     */
    bool isAnyType() const;

  /* ************************************************************************ */
  protected:

    /** The runtime we belong to, for reference to other types and the like */
    const Runtime& rt;

  /* ************************************************************************ */
  private:

    /** The atom naming this type, needed for message outputs */
    const Atom name; 

    /** Vector-type used to store the super-types */
    typedef std::vector<Ref> supersT;
    /** Array of super-types */
    supersT supers;

};

} // namespace ellisp
} // namespace lispel

#endif // Header-guard
