/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "UserAction.hpp"
#include "Runtime.hpp"
#include "List.hpp"
#include "Cell.hpp"

#include <cassert>
#include <stdexcept>

namespace lispel
{
namespace ellisp
{

#define THROW(msgout) \
  { \
    std::ostringstream msg; \
    msg << msgout; \
    throw std::runtime_error(msg.str()); \
  }

/* ************************************************************************** */
/* ****************************** UserAction ******************************** */
/* ************************************************************************** */

#define NAME_ACTION Atom("_action")

UserAction::UserAction(const Runtime& r, List* lst)
  : SignatedAction(r), name(), argNames()
#ifndef NDEBUG
  , code(NULL)
#endif
{
  // Must look like action definition
  assert(isAction(lst));
  assert(lst);
  assert(lst->getCar().getType()==Cell::ATOM);
  assert(lst->getCar().getAtom()==NAME_ACTION);

  // Next comes action's name
  lst=lst->getCdr();
  if(!lst)
    throw std::runtime_error("Expected action name!");
  if(lst->getCar().getType()!=Cell::ATOM)
    throw std::runtime_error("Action-name must be Atom!");
  name=lst->getCar().getAtom();

  // Then arguments
  lst=lst->getCdr();
  if(!lst)
    THROW("Missing signature for action " << name)
  if(lst->getCar().getType()!=Cell::LIST)
    THROW("Expected list for signature of action " << name)
  Signature::parse(rt, lst->getCar().getList(), arguments, argNames);

  // And finally the code
  code=lst->spliceCdr();
  assert(!lst->getCdr());
  if(!code)
    THROW("Missing action-code for " << name)
  if(code->getCdr())
  {
    delete code; // Our destructor will not be called for now
    THROW("Did not expect cell after code of function " << name)
  }

  // Check the code's validity
  if(code->getCar().getType()!=Cell::LIST)
    THROW("Code of action " << name << " should be list!")
  if(!code->getCar().getList())
    THROW("Code of action " << name << " must not be empty!")
  for(const List* i(code->getCar().getList()); i; i=i->getCdr())
    if(i->getCar().getType()!=Cell::LIST)
      THROW("Code of action " << name << " must only consist of Lists!")
}

UserAction::~UserAction()
{
  assert(code);
  delete code;
}

void UserAction::define(Runtime& r)
{
  assert(static_cast<const Runtime*>(&r)==&rt);
  r.addAction(name, this);
}

bool UserAction::isAction(const List* lst)
{
  // It mustn't be empty
  if(!lst)
    return false;

  // First element should be Atom _action
  if(lst->getCar().getType()!=Cell::ATOM)
    return false;
  if(lst->getCar().getAtom()!=NAME_ACTION)
    return false;

  return true;
}

void UserAction::performCasted(Arguments& vals) const
{
  // Open new scope and define arguments
  Runtime execScope(rt);
  arguments.defineArgs(execScope, argNames, vals);

  // Evaluate the user-code command after command
  assert(code);
  assert(code->getCar().getList());
  for(const List* cmd(code->getCar().getList()); cmd; cmd=cmd->getCdr())
  {
    assert(cmd->getCar().getType()==Cell::LIST); // Checked in constructor
    execScope.processActionBody(cmd->getCar().getList());
  }
}

} // namespace ellisp
} // namespace lispel
