/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "UserClass.hpp"
#include "Elob.hpp"
#include "Runtime.hpp"
#include "List.hpp"
#include "Cell.hpp"

#include <stdexcept>
#include <sstream>
#include <algorithm>

namespace lispel
{
namespace ellisp
{

#define NAME_CLASS Atom("_class")

Atom UserClass::getAccessorName(const Atom& m) const
{
  return Atom(getName().str()+"-"+m.str());
}

Atom UserClass::getConstructorName() const
{
  return Atom("Neu-"+getName().str());
}

#define THROW(out) \
  { \
    std::ostringstream msg; \
    msg << out; \
    throw std::runtime_error(msg.str()); \
  }

/* ************************************************************************** */
/* ****************************** UserClass ********************************* */
/* ************************************************************************** */

UserClass::UserClass(const Runtime& r, const List* lst)
  : Type(getClassName(lst), r), members()
{
  // Must be class, skip keyword
  assert(isClass(lst));
  assert(lst);
  
  // Name already defined!
  lst=lst->getCdr();
  assert(lst);
  assert(lst->getCar().getType()==Cell::ATOM);
  assert(lst->getCar().getAtom()==getName());

  // Next is list of supers
  lst=lst->getCdr();
  if(!lst)
    THROW("Class " << getName() << " is missing list of super-classes!")
  if(lst->getCar().getType()!=Cell::LIST)
    THROW("List of super-classes of " << getName() << " must be a List!")
  parseMemberList(lst->getCar().getList(), true);

  // Then list of members
  lst=lst->getCdr();
  if(!lst)
    THROW("Class " << getName() << " is missing list of members!")
  if(lst->getCar().getType()!=Cell::LIST)
    THROW("List of members of " << getName() << " must be a List!")
  parseMemberList(lst->getCar().getList(), false);

  // And nothing more!
  if(lst->getCdr())
    THROW("Too many elements for definition of class " << getName())

  // Now, register super-types
  for(memberArr::const_iterator i(members.begin()); i!=members.end(); ++i)
    if(i->isSuper)
      addSuper(i->type);
}

void UserClass::parseMemberList(const List* lst, bool isSuper)
{
  for(; lst; lst=lst->getCdr())
  {
    if(lst->getCar().getType()!=Cell::LIST)
      THROW("Expected list as member-definition in " << getName())
    const List* memLst(lst->getCar().getList());

    memberEntry mem;
    mem.isSuper=isSuper;

    // Next comes type
    if(!memLst)
      THROW("Member-definition in " << getName() << " misses type")
    if(memLst->getCar().getType()!=Cell::ATOM)
      THROW("Member-type in " << getName() << " must be Atom")
    mem.type=rt.queryType(memLst->getCar().getAtom());

    // And possibly name
    memLst=memLst->getCdr();
    mem.hasName=(!isSuper || memLst); // Should we have a name?
    if(mem.hasName)
    {
      if(!memLst)
        THROW("Member-definition in " << getName() << " misses name")
      if(memLst->getCar().getType()!=Cell::ATOM)
        THROW("Member-name in " << getName() << " must be Atom")
      mem.name=memLst->getCar().getAtom();
      memLst=memLst->getCdr();
    }

    // Nothing more
    if(memLst)
      THROW("Too many elements in member-definition in " << getName())

    // Save it
    members.push_back(mem);
  }
}

const Atom& UserClass::getClassName(const List* lst)
{
  // Skip keyword
  assert(isClass(lst));
  assert(lst);
  lst=lst->getCdr();
  assert(lst);

  // Next is name
  if(!lst)
    throw std::runtime_error("Class-definition missing class name!");
  if(lst->getCar().getType()!=Cell::ATOM)
    throw std::runtime_error("Class name must be Atom!");
  return lst->getCar().getAtom();
}

void UserClass::define(Runtime& r)
{
  assert(const_cast<const Runtime*>(&r)==&rt);
  r.addType(getName(), this);

  // Accessor methods, all have signature xxx(myClass obj)
  Signature sig;
  sig.push_back(ref());
  size_t j(0);
  for(memberArr::const_iterator i(members.begin()); i!=members.end(); ++i, ++j)
    if(i->hasName)
      r.addFunction(getAccessorName(i->name), new AccessorFunction(r, sig, j));

  // Constructor-method
  sig.clear();
  for(memberArr::const_iterator i(members.begin()); i!=members.end(); ++i)
    sig.push_back(i->type);
  r.addFunction(getConstructorName(), new ConstructorFunction(r, sig, ref()));
}

void UserClass::destroyData(void* val) const
{
  delete[] static_cast<Elob*>(val);
}

Elob UserClass::castValueTo(const Elob& val, Ref t) const
{
  assert(val.getType()==ref());

  // Cast must be valid, implying non-ambiguousness
  assert(t->isAssignableFrom(ref()));

  // Now find correct super-type for this cast
  const Elob* vals(static_cast<const Elob*>(val.getValue()));
  for(memberArr::const_iterator i(members.begin()); i!=members.end();
    ++i, ++vals)
  {
    if(i->isSuper && t->isAssignableFrom(i->type))
    {
      // Cast to this super type
      const Elob& curMem(*vals);
      assert(i->type==curMem.getType());
      Elob casted(i->type, const_cast<void*>(curMem.getValue()), val);

      // Return if exact match, otherwise recurse
      if(t==i->type)
        return casted;
      return i->type->castValueTo(casted, t);
    }
  }

  // Nothing found!
  assert(false);
}

std::ostream& UserClass::print(std::ostream& out, const void* val) const
{
  const Elob* vals(static_cast<const Elob*>(val));

  out << getName() << std::endl << "{" << std::endl;
  for(memberArr::const_iterator i(members.begin()); i!=members.end();
    ++i, ++vals)
  {
    out << " ";
    if(i->hasName)
      out << i->name << "=";
    out << *vals;
    if(i->isSuper)
      out << " (*)";
    out << std::endl;
  }
  out << "}" << std::endl;

  return out;
}

bool UserClass::isClass(const List* lst)
{
  // Must be (_class ...)
  if(!lst)
    return false;
  if(lst->getCar().getType()!=Cell::ATOM)
    return false;
  if(lst->getCar().getAtom()!=NAME_CLASS)
    return false;

  return true;
}

/* ************************************************************************** */
/* ************************** UserClass::AccessorFunction ******************* */
/* ************************************************************************** */

Elob UserClass::AccessorFunction::evalCasted(const Arguments& vals) const
{
  assert(vals.size()==1);
  const Elob* obj(static_cast<const Elob*>(vals[0].getValue()));
  return obj[memberInd];
}

/* ************************************************************************** */
/* *********************** UserClass::ConstructorFunction ******************* */
/* ************************************************************************** */

Elob UserClass::ConstructorFunction::evalCasted(const Arguments& vals) const
{
  // Build up the array, assuming nothing is thrown here!
  Elob* obj(new Elob[vals.size()]);
  std::copy(vals.begin(), vals.end(), obj);

  // Return, takes ownership of array
  return Elob(resType, obj, rt);
}

} // namespace ellisp
} // namespace lispel
