/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef ELLISP_USERCLASS_HPP
#define ELLISP_USERCLASS_HPP

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "Type.hpp"
#include "Function.hpp"
#include "Atom.hpp"

#include <iostream>
#include <vector>
#include <utility>
#include <cstddef>

namespace lispel
{
namespace ellisp
{

class Runtime;
class Elob;
class List;

/* ************************************************************************** */
/* ****************************** UserClass ********************************* */
/* ************************************************************************** */

/**
 * A user-defined type / "class"; objects of such types are simply a collection
 * of sub-Elobs (much like struct's).  Some of them are defined their
 * "super-objects" and they may be casted to them, and some are their members,
 * which can be accessed by accessor-methods.
 *
 * The data-value is simply an array of Elob's in the layout:
 * [super object 1]
 * [super object 2]
 * ...
 * [super object m]
 * [member 1]
 * [member 2]
 * ...
 * [member n]
 */
class UserClass : public Type
{

  /* ************************************************************************ */
  public:

    /**
     * Create a new UserClass by its definition in List-form.
     * @param r The runtime
     * @param lst The List to extract from
     */
    UserClass(const Runtime& r, const List* lst);

    // No destructor needed

    /**
     * Define the class in the given Runtime, should be non-const version
     * of the real runtime.  This not only defines the type but also the
     * accessor- and constructor-functions and the like.
     * @param r The runtime to define in
     */
    void define(Runtime& r);

    // Override type's methods
    virtual void destroyData(void* val) const;
    virtual Elob castValueTo(const Elob& val, Ref t) const;
    virtual std::ostream& print(std::ostream& out, const void* val) const;

    /**
     * Check if a list describes a user-class-definition.
     * @param lst The List
     * @return True if it does
     */
    static bool isClass(const List* lst);

  /* ************************************************************************ */
  private:

    /** Function-class used for the accessor-functions */
    class AccessorFunction;
    /** Function used as constructor */
    class ConstructorFunction;

    /** Type for entry in "member" (including supers) list */
    struct memberEntry
    {
      /** This member's type */
      Type::Ref type;
      /** This member's name, if it has */
      Atom name;
      /** If this member has a name */
      bool hasName;
      /** If it is a super-type (we are castable to it) */
      bool isSuper;
    };

    /** Type of array for member-entries */
    typedef std::vector<memberEntry> memberArr;
    /** Array of member entries */
    memberArr members;

    /**
     * Parse a member-definition list, either for supers or members.
     * @param lst The List to parse
     * @param isSuper Whether it describes a list of supers
     */
    void parseMemberList(const List* lst, bool isSuper);

    // Build accessor-/constructor-function name
    Atom getAccessorName(const Atom& m) const;
    Atom getConstructorName() const;

    /**
     * Helper-method to extract the class-name out of a List, so we have
     * access to it for Type's constructor in our constructor.
     * @param lst The List representing this class
     * @return The name of the class according to the definition
     */
    static const Atom& getClassName(const List* lst);

};

/* ************************************************************************** */
/* ************************** UserClass::AccessorFunction ******************* */
/* ************************************************************************** */

/**
 * Class used to implement and define the accessor-functions to members of an
 * UserClass-object
 */
class UserClass::AccessorFunction : public SignatedFunction
{

  /* ************************************************************************ */
  public:

    /**
     * Construct it.
     * @param r The Runtime
     * @param sig The signature
     * @param ind The desired member's index
     */
    inline AccessorFunction(const Runtime& r, const Signature& sig, size_t ind)
      : SignatedFunction(r, sig), memberInd(ind)
    {}

  /* ************************************************************************ */
  protected:

    /**
     * @see SignatedFunction::evalCasted(const Arguments& vals) const
     */
    virtual Elob evalCasted(const Arguments& vals) const;

  /* ************************************************************************ */
  private:

    /** Index of the desired member inside the object, supers counted */
    const size_t memberInd;

};

/* ************************************************************************** */
/* *********************** UserClass::ConstructorFunction ******************* */
/* ************************************************************************** */

/**
 * Class used to implement and define the constructor-function for an UserClass
 * object; it simply creates the Elob-array for all its arguments, since the
 * signature makes sure they match.
 */
class UserClass::ConstructorFunction : public SignatedFunction
{

  /* ************************************************************************ */
  public:

    /**
     * Construct it.
     * @param r The Runtime
     * @param sig The signature
     * @param resT The type to assign the resulting object
     */
    inline ConstructorFunction(const Runtime& r, const Signature& sig,
      Type::Ref resT)
      : SignatedFunction(r, sig), resType(resT)
    {}

  /* ************************************************************************ */
  protected:

    /**
     * @see SignatedFunction::evalCasted(const Arguments& vals) const
     */
    virtual Elob evalCasted(const Arguments& vals) const;

  /* ************************************************************************ */
  private:

    /** Type to assign the resulting object */
    const Type::Ref resType;

};

} // namespace ellisp
} // namespace lispel

#endif // Header-guard
