/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "UserFunction.hpp"
#include "Atom.hpp"
#include "Runtime.hpp"
#include "List.hpp"
#include "Cell.hpp"
#include "Arguments.hpp"

#include <stdexcept>
#include <sstream>
#include <cassert>

namespace lispel
{
namespace ellisp
{

#define NAME_FUNCTION Atom("_function")

/* ************************************************************************** */
/* ****************************** UserFunction ****************************** */
/* ************************************************************************** */

UserFunction::UserFunction(const Runtime& r, List* lst)
  : SignatedFunction(r), name(), argNames()
#ifndef NDEBUG
  , code(NULL)
#endif
{
  // Must look like function definition
  assert(isFunction(lst));
  assert(lst);
  assert(lst->getCar().getType()==Cell::ATOM);
  assert(lst->getCar().getAtom()==NAME_FUNCTION);

  // Next comes function's name
  lst=lst->getCdr();
  if(!lst)
    throw std::runtime_error("Expected function name!");
  if(lst->getCar().getType()!=Cell::ATOM)
    throw std::runtime_error("Function-name must be Atom!");
  name=lst->getCar().getAtom();

  // Then arguments
  lst=lst->getCdr();
  if(!lst)
  {
    std::ostringstream msg;
    msg << "Missing signature for function " << name;
    throw std::runtime_error(msg.str());
  }
  if(lst->getCar().getType()!=Cell::LIST)
  {
    std::ostringstream msg;
    msg << "Expected list for signature of function " << name;
    throw std::runtime_error(msg.str());
  }
  Signature::parse(rt, lst->getCar().getList(), arguments, argNames);

  // And finally the code
  code=lst->spliceCdr();
  assert(!lst->getCdr());
  if(!code)
  {
    std::ostringstream msg;
    msg << "Missing function-code for " << name;
    throw std::runtime_error(msg.str());
  }
  if(code->getCdr())
  {
    delete code; // Our destructor will not be called for now
    std::ostringstream msg;
    msg << "Did not expect cell after code of function " << name;
    throw std::runtime_error(msg.str());
  }
}

UserFunction::~UserFunction()
{
  assert(code);
  delete code;
}

void UserFunction::define(Runtime& r)
{
  assert(static_cast<const Runtime*>(&r)==&rt);
  r.addFunction(name, this);
}

bool UserFunction::isFunction(const List* lst)
{
  // No empty list!
  if(!lst)
    return false;

  // Must start with keyword
  if(lst->getCar().getType()!=Cell::ATOM)
    return false;
  if(lst->getCar().getAtom()!=NAME_FUNCTION)
    return false;

  // Otherwise, it should be a function definition; validity is checked in c'tor
  return true;
}

Elob UserFunction::evalCasted(const Arguments& vals) const
{
  // Open new scope and define arguments
  Runtime execScope(rt);
  arguments.defineArgs(execScope, argNames, vals);

  // Evaluate the user-code to get the result
  assert(code);
  return execScope.evaluate(code->getCar());
}

} // namespace ellisp
} // namespace lispel
