/*

    Lispel, typed German language Lisp.
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

/* Tests the stream API; basically first writes to a file and then checks to see
 * if the data read from it again is correct.
 */

#include "Test.hpp"
#include "ClassicEval.hpp"

#include "Conditions.hpp"
#include "Comparisons.hpp"
#include "Variables.hpp"
#include "Streams.hpp"

#include <cstdlib>

using namespace lispel;
using namespace ellisp;
using namespace ellib;
using namespace tests;

using namespace std;

#define TEMP_FILE "streams.tmp"

int main()
{
  ClassicEval proc;
  Conditions::define(proc);
  Comparisons::define(proc);
  Variables::define(proc);
  Streams::define(proc);

  Elob val;

  // Define a variable used for data-transfer
  proc.procClassic("(_constant x (Variable 0))");
  val=proc.evalClassic("x", Variables::getVariableType(proc));
  Elob& var(*static_cast<Elob*>(val.getValue()));

  // Variable set to the current stream
  proc.procClassic("(_constant stream (Variable 0))");

  /* ********** Binary-File-IO ********** */

  // Open the stream
  proc.procClassic("(setze stream (Binaerdatei-Outputstream \""
    TEMP_FILE "\"))");
  
  // Write out the bytes 0..255
  for(int b(0); b!=0x100; ++b)
  {
    var=Elob::forInteger(b, proc);
    proc.procClassic("(schreibe-Byte (Wert stream) (Wert x))");
  }

  // Open the file back for reading
  proc.procClassic("(setze stream 0)"); /* Destroy and close the open file */
  proc.procClassic("(setze stream (Binaerdatei-Inputstream \""
    TEMP_FILE "\"))");

  // Read in the bytes again
  for(int b(0); b!=0x100; ++b)
  {
    proc.procClassic("(lese-Byte (Wert stream) x)");
    check(var.getType()==proc.getIntegerType(), 
      "readByte returned wrong type!");
    check(var.getInteger()==b, "readByte returned wrong value!");
  }

  // Now there should be EOF
  proc.procClassic("(lese-Byte (Wert stream) x)");
  val=proc.evalClassic("(Gleichheit EOF (Wert x))", proc.getBooleanType());
  check(val.getBoolean(), "Did not get EOF as expected!");

  /* ********** Text-File-IO ********** */

  // Open the stream
  proc.procClassic("(setze stream (Textdatei-Outputstream \""
    TEMP_FILE "\"))");

  // Write out two lines
  proc.procClassic("(schreibe (Wert stream) \"6*7=42\nabc\n\")");

  // Open the file back for reading
  proc.procClassic("(setze stream 0)"); /* Destroy and close the open file */
  proc.procClassic("(setze stream (Textdatei-Inputstream \""
    TEMP_FILE "\"))");

  // Read the lines and check
  proc.procClassic("(lese-Zeile (Wert stream) x)");
  check(var.getType()==proc.getStringType(), "readLine returned wrong type!");
  check(var.getString()=="6*7=42", "Line read is wrong (%s)!",
    var.getString().c_str());
  proc.procClassic("(lese-Zeile (Wert stream) x)");
  check(var.getString()=="abc", "Second line read is wrong!");

  /* ********** String-Input ********** */

  // Open the stream
  proc.procClassic("(setze stream (String-Inputstream "
    "\"abcde 5 .25 \\\"d\\\"\"))");

  // Read in a sub-string by length
  proc.procClassic("(lese-String-nach-Laenge (Wert stream) 3 x)");
  check(var.getType()==proc.getStringType(), 
    "read by length returned wrong type!");
  check(var.getString()=="abc", "read by length returned wrong value (%s)!",
    var.getString().c_str());
  proc.procClassic("(lese-String-nach-Laenge (Wert stream) 2 x)");

  // Read in with input
  #define CHECK_INPUT(type, value) \
    proc.procClassic("(lese-ein (Wert stream) x)"); \
    check(var.getType()==proc.get##type##Type(), \
      "Input of " #type " wrong type!"); \
    check(var.get##type()==value, "Input of " #type " wrong value!");
  CHECK_INPUT(Integer, 5)
  CHECK_INPUT(Number, .25)
  CHECK_INPUT(String, "d")


  /* ************ Done ************* */

  // Just for fun: Print out value to stdout
  proc.procClassic("(gib-aus stdout x)");
  proc.procClassic("(schreibe stdout \"\n\")");

  return EXIT_SUCCESS;
}
